"""This module contains the engine class that will run the complete
timeseries simulations."""

import numpy as np
from pvfactors.viewfactors import VFCalculator
from pvfactors.irradiance import HybridPerezOrdered
from pvfactors.config import DEFAULT_RHO_FRONT, DEFAULT_RHO_BACK


class PVEngine(object):
    """Class putting all of the calculations together into simple
    workflows.
    """

    def __init__(self, pvarray, vf_calculator=None, irradiance_model=None,
                 fast_mode_pvrow_index=None, fast_mode_segment_index=None):
        """Create pv engine class, and initialize timeseries parameters.

        Parameters
        ----------
        pvarray : BasePVArray (or child) object
            The initialized PV array object that will be used for calculations
        vf_calculator : vf calculator object, optional
            Calculator that will be used to calculate the view factor matrices,
            will use :py:class:`~pvfactors.viewfactors.calculator.VFCalculator`
            if None (Default = None)
        irradiance_model : irradiance model object, optional
            The irradiance model that will be applied to the PV array,
            will use
            :py:class:`~pvfactors.irradiance.models.HybridPerezOrdered`
            if None (Default = None)
        fast_mode_pvrow_index : int, optional
            If a pvrow index is passed, then the PVEngine fast mode
            will be activated and the engine calculation will be done only
            for the back surface of the pvrow with the corresponding
            index (Default = None)
        fast_mode_segment_index : int, optional
            If a segment index is passed, then the PVEngine fast mode
            will calculate back surface irradiance only for the
            selected segment of the selected back surface (Default = None)
        """
        # Initialize the attributes of the PV engine
        self.vf_calculator = vf_calculator or VFCalculator()
        self.irradiance = irradiance_model or HybridPerezOrdered()
        self.pvarray = pvarray
        # Save fast mode indices
        self.fast_mode_pvrow_index = fast_mode_pvrow_index
        self.fast_mode_segment_index = fast_mode_segment_index

        # These values will be updated at fitting time
        self.n_points = None
        self.skip_step = None

    @classmethod
    def with_rho_initialization(
            cls, pvarray, vf_calculator, irradiance_model,
            fast_mode_pvrow_index=None, fast_mode_segment_index=None):
        """Before creating the PV engine object, update the front and
        back reflectivity scalars using the faoi functions, if those values
        weren't passed originally

        Parameters
        ----------
        pvarray : :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`
            The initialized PV array object that will be used for calculations
        vf_calculator : \
        :py:class:`~pvfactors.viewfactors.calculator.VFCalculator`
            Calculator that will be used to calculate the view factor matrices,
            and AOI losses
        irradiance_model : irradiance model object
            The irradiance model that will be applied to the PV array,
            for instance
            :py:class:`~pvfactors.irradiance.models.HybridPerezOrdered`
        fast_mode_pvrow_index : int, optional
            If a pvrow index is passed, then the PVEngine fast mode
            will be activated and the engine calculation will be done only
            for the back surface of the pvrow with the corresponding
            index (Default = None)
        fast_mode_segment_index : int, optional
            If a segment index is passed, then the PVEngine fast mode
            will calculate back surface irradiance only for the
            selected segment of the selected back surface (Default = None)

        Returns
        -------
        PV engine
            PV engine where the rho values have been initialized
        """
        # Calculate global average reflectivity using VF calculator
        is_back = False
        rho_front_calculated = (vf_calculator.vf_aoi_methods
                                .rho_from_faoi_fn(is_back))
        is_back = True
        rho_back_calculated = (vf_calculator.vf_aoi_methods
                               .rho_from_faoi_fn(is_back))
        # Initialize rho values for irradiance model
        irradiance_model.rho_front = irradiance_model.initialize_rho(
            irradiance_model.rho_front, rho_front_calculated,
            DEFAULT_RHO_FRONT)
        irradiance_model.rho_back = irradiance_model.initialize_rho(
            irradiance_model.rho_back, rho_back_calculated,
            DEFAULT_RHO_BACK)

        return cls(pvarray, vf_calculator=vf_calculator,
                   irradiance_model=irradiance_model,
                   fast_mode_pvrow_index=fast_mode_pvrow_index,
                   fast_mode_segment_index=fast_mode_segment_index)

    def fit(self, timestamps, DNI, DHI, solar_zenith, solar_azimuth,
            surface_tilt, surface_azimuth, albedo, ghi=None):
        """Fit the timeseries data to the engine. More specifically,
        save all the parameters that needs to be saved, and fit the PV array
        and irradiance models to the data (i.e. perform all the intermediate
        vector-based calculations).
        Note that all angles follow the pvlib-python angle convention: North -
        0 deg, East - 90 deg, etc.

        Parameters
        ----------
        timestamps : array-like or timestamp-like
            List of timestamps of the simulation.
        DNI : array-like or float
            Direct normal irradiance values [W/m2]
        DHI : array-like or float
            Diffuse horizontal irradiance values [W/m2]
        solar_zenith : array-like or float
            Solar zenith angles [deg]
        solar_azimuth : array-like or float
            Solar azimuth angles [deg]
        surface_tilt : array-like or float
            Surface tilt angles, from 0 to 180 [deg]
        surface_azimuth : array-like or float
            Surface azimuth angles [deg]
        albedo : array-like or float
            Albedo values (ground reflectivity)
        ghi : array-like, optional
            Global horizontal irradiance [W/m2], if not provided, will be
            calculated from DNI and DHI if needed (Default = None)
        """
        # Format inputs to numpy arrays if it looks like floats where inputted
        if np.isscalar(DNI):
            timestamps = [timestamps]
            DNI = np.array([DNI])
            DHI = np.array([DHI])
            solar_zenith = np.array([solar_zenith])
            solar_azimuth = np.array([solar_azimuth])
            surface_tilt = np.array([surface_tilt])
            surface_azimuth = np.array([surface_azimuth])
            ghi = None if ghi is None else np.array([ghi])

        # Format albedo
        self.n_points = len(DNI)
        albedo = (albedo * np.ones(self.n_points) if np.isscalar(albedo)
                  else albedo)

        # Fit PV array
        self.pvarray.fit(solar_zenith, solar_azimuth, surface_tilt,
                         surface_azimuth)

        # Fit irradiance model
        self.irradiance.fit(timestamps, DNI, DHI, solar_zenith, solar_azimuth,
                            surface_tilt, surface_azimuth, albedo, ghi=ghi)

        # Add timeseries irradiance results to pvarray
        self.irradiance.transform(self.pvarray)

        # Fit VF calculator
        self.vf_calculator.fit(self.n_points)

        # Skip timesteps when:
        #    - solar zenith > 90, ie the sun is down
        #    - DNI or DHI is negative, which does not make sense
        #    - DNI and DHI are both zero
        self.skip_step = (solar_zenith > 90) | (DNI < 0) | (DHI < 0) \
            | ((DNI == 0) & (DHI == 0))

    def run_full_mode(self, fn_build_report=None):
        """Run all simulation timesteps using the full mode, which calculates
        the equilibrium of reflections in the system, and returns a report that
        will be built by the function passed by the user.

        Parameters
        ----------
        fn_build_report : function, optional
            Function that will build the report of the simulation
            (Default value = None)

        Returns
        -------
        report
            Saved results from the simulation, as specified by user's report
            function. If no function is passed, nothing will be returned.
        """
        # Get pvarray
        pvarray = self.pvarray

        # Get the irradiance modeling matrices
        # shape = n_surfaces + 1, n_timesteps
        irradiance_mat, rho_mat, invrho_mat, _ = \
            self.irradiance.get_full_ts_modeling_vectors(pvarray)

        # --- Calculate view factors
        # shape = n_surfaces + 1, n_surfaces + 1, n_timesteps
        ts_vf_matrix = self.vf_calculator.build_ts_vf_matrix(pvarray)
        pvarray.ts_vf_matrix = ts_vf_matrix
        # Reshape for broadcasting and inverting
        # shape = n_timesteps, n_surfaces + 1, n_surfaces + 1
        ts_vf_matrix_reshaped = np.moveaxis(ts_vf_matrix, -1, 0)

        # --- Solve mathematical problem
        # shape of system = n_timesteps, n_surfaces + 1, n_surfaces + 1
        shape_system = ts_vf_matrix_reshaped.shape
        # Build 3d matrix of inverse reflectivities: diagonal for each ts
        # shape = n_timesteps, n_surfaces + 1, n_surfaces + 1
        invrho_ts_diag = np.zeros(shape_system)
        for idx_surf in range(shape_system[1]):
            invrho_ts_diag[:, idx_surf, idx_surf] = invrho_mat[idx_surf, :]
        # Subtract matrices: will rely on broadcasting
        # shape = n_timesteps, n_surfaces + 1, n_surfaces + 1
        a_mat = invrho_ts_diag - ts_vf_matrix_reshaped
        # Calculate inverse, requires specific shape
        # shape = n_timesteps, n_surfaces + 1, n_surfaces + 1
        inv_a_mat = np.linalg.inv(a_mat)
        # Use einstein sum to get final timeseries radiosities
        # shape = n_surfaces + 1, n_timesteps
        q0 = np.einsum('ijk,ki->ji', inv_a_mat, irradiance_mat)
        # Calculate incident irradiance: will rely on broadcasting
        # shape = n_surfaces + 1, n_timesteps
        qinc = np.einsum('ijk,ki->ji', invrho_ts_diag, q0)

        # --- Derive other irradiance terms
        # shape = n_surfaces, n_timesteps
        isotropic_mat = ts_vf_matrix[:-1, -1, :] * irradiance_mat[-1, :]
        reflection_mat = qinc[:-1, :] - irradiance_mat[:-1, :] - isotropic_mat

        # --- Calculate AOI losses and absorbed irradiance
        # Create tiled reflection matrix of
        # shape = n_surfaces + 1, n_surfaces + 1, n_timestamps
        rho_ts_tiled = np.moveaxis(np.tile(rho_mat.T, (shape_system[1], 1, 1)),
                                   -1, 0)
        # Get vf AOI matrix
        # shape [n_surfaces + 1, n_surfaces + 1, n_timestamps]
        vf_aoi_matrix = (self.vf_calculator
                         .build_ts_vf_aoi_matrix(pvarray, rho_ts_tiled))
        pvarray.ts_vf_aoi_matrix = vf_aoi_matrix
        # Get absorbed irradiance matrix
        # shape [n_surfaces, n_timestamps]
        irradiance_abs_mat = (
            self.irradiance.get_summed_components(pvarray, absorbed=True))
        # Calculate absorbed irradiance
        qabs = (np.einsum('ijk,jk->ik', vf_aoi_matrix, q0)[:-1, :]
                + irradiance_abs_mat)

        # --- Update surfaces with values: the lists are ordered by index
        for idx_surf, ts_surface in enumerate(pvarray.all_ts_surfaces):
            ts_surface.update_params(
                {'q0': q0[idx_surf, :],
                 'qinc': qinc[idx_surf, :],
                 'isotropic': isotropic_mat[idx_surf, :],
                 'reflection': reflection_mat[idx_surf, :],
                 'qabs': qabs[idx_surf, :]})

        # Return report if function was passed
        report = None if fn_build_report is None else fn_build_report(pvarray)
        return report

    def run_fast_mode(self, fn_build_report=None, pvrow_index=None,
                      segment_index=None):
        """Run all simulation timesteps using the fast mode for the back
        surface of a PV row, and assuming that the incident irradiance on all
        other surfaces is known (all but back surfaces).
        The function will return a report that will be built by the function
        passed by the user.

        Parameters
        ----------
        fn_build_report : function, optional
            Function that will build the report of the simulation
            (Default value = None)
        pvrow_index : int, optional
            Index of the PV row for which we want to calculate back surface
            irradiance; if used, this will override the
            ``fast_mode_pvrow_index`` passed at engine initialization
            (Default = None)
        segment_index : int, optional
            Index of the segment on the PV row's back side for which we want
            to calculate the incident irradiance; if used, this will override
            the ``fast_mode_segment_index`` passed at engine initialization
            (Default = None)

        Returns
        -------
        report
            Results from the simulation, as specified by user's report
            function. If no function is passed, nothing will be returned.
        """

        # Prepare variables
        pvrow_idx = (self.fast_mode_pvrow_index if pvrow_index is None
                     else pvrow_index)
        segment_idx = (self.fast_mode_segment_index if segment_index is None
                       else segment_index)
        ts_pvrow = self.pvarray.ts_pvrows[pvrow_idx]

        # Run calculations
        if segment_idx is None:
            # Run calculation for all segments of back surface
            for ts_segment in ts_pvrow.back.list_segments:
                self._calculate_back_ts_segment_qinc(ts_segment, pvrow_idx)
        else:
            # Run calculation for selected segment of back surface
            ts_segment = ts_pvrow.back.list_segments[segment_idx]
            self._calculate_back_ts_segment_qinc(ts_segment, pvrow_idx)

        # Create report
        report = (None if fn_build_report is None
                  else fn_build_report(self.pvarray))

        return report

    def _calculate_back_ts_segment_qinc(self, ts_segment, pvrow_idx):
        """Calculate the incident irradiance on a timeseries segment's surfaces
        for the back side of a PV row, using the fast mode, so assuming that
        the incident irradiance on all other surfaces is known.
        Nothing is returned by the function, but the segment's surfaces'
        parameter values are updated.

        Parameters
        ----------
        ts_segment : :py:class:`~pvfactors.geometry.timeseries.TsDualSegment`
            Timeseries segment for which we want to calculate the incident
            irradiance
        pvrow_idx : int
            Index of the PV row on which the segment is located
        """

        # Get all timeseries surfaces in segment
        list_ts_surfaces = ts_segment.all_ts_surfaces

        # Get irradiance vectors for calculation
        albedo = self.irradiance.albedo
        rho_front = self.irradiance.rho_front
        irr_gnd_shaded = self.irradiance.gnd_shaded
        irr_gnd_illum = self.irradiance.gnd_illum
        irr_pvrow_shaded = self.irradiance.pvrow_shaded
        irr_pvrow_illum = self.irradiance.pvrow_illum
        irr_sky = self.irradiance.sky_luminance

        for ts_surface in list_ts_surfaces:
            # Calculate view factors for timeseries surface
            vf = self.vf_calculator.get_vf_ts_pvrow_element(
                pvrow_idx, ts_surface, self.pvarray.ts_pvrows,
                self.pvarray.ts_ground, self.pvarray.rotation_vec,
                self.pvarray.width)

            # Update sky terms of timeseries surface
            self.irradiance.update_ts_surface_sky_term(ts_surface)

            # Calculate incident irradiance on illuminated surface
            gnd_shadow_refl = vf['to_gnd_shaded'] * albedo * irr_gnd_shaded
            gnd_illum_refl = vf['to_gnd_illum'] * albedo * irr_gnd_illum
            pvrow_shadow_refl = (vf['to_pvrow_shaded'] * rho_front
                                 * irr_pvrow_shaded)
            pvrow_illum_refl = (vf['to_pvrow_illum'] * rho_front
                                * irr_pvrow_illum)
            reflections = (gnd_shadow_refl + gnd_illum_refl + pvrow_shadow_refl
                           + pvrow_illum_refl)
            isotropic = vf['to_sky'] * irr_sky
            qinc = (gnd_shadow_refl + gnd_illum_refl + pvrow_shadow_refl
                    + pvrow_illum_refl + isotropic
                    + ts_surface.get_param('sky_term'))

            # Update parameters of timeseries surface object
            ts_surface.update_params(
                {'qinc': qinc,
                 'reflection_gnd_shaded': gnd_shadow_refl,
                 'reflection_gnd_illum': gnd_illum_refl,
                 'reflection_pvrow_shaded': pvrow_shadow_refl,
                 'reflection_pvrow_illum': pvrow_illum_refl,
                 'isotropic': isotropic,
                 'reflection': reflections,
                 'view_factors': vf})



"""Module containing PV array classes, which will use PV rows and ground
geometries."""

import numpy as np
from pvfactors.config import X_ORIGIN_PVROWS, DISTANCE_TOLERANCE
from pvfactors.geometry.base import \
    _get_solar_2d_vectors, BasePVArray, _get_rotation_from_tilt_azimuth
from pvfactors.geometry.pvrow import TsPVRow
from pvfactors.geometry.pvground import TsGround


class OrderedPVArray(BasePVArray):
    """An ordered PV array has a flat horizontal ground, and pv rows which
    are all at the same height, with the same surface tilt and azimuth angles,
    and also all equally spaced. These simplifications allow faster and easier
    calculations. In the ordered PV array, the list of PV rows must be
    ordered from left to right (along the x-axis) in the 2D geometry."""

    y_ground = 0.  # ground will be at height = 0 by default

    def __init__(self, axis_azimuth=None, gcr=None, pvrow_height=None,
                 n_pvrows=None, pvrow_width=None, param_names=None,
                 cut=None):
        """Initialize ordered PV array.
        List of PV rows will be ordered from left to right.

        Parameters
        ----------
        axis_azimuth : float, optional
            Azimuth angle of rotation axis [deg] (Default = None)
        gcr : float, optional
            Ground coverage ratio (Default = None)
        pvrow_height : float, optional
            Unique height of all PV rows in [m] (Default = None)
        n_pvrows : int, optional
            Number of PV rows in the PV array (Default = None)
        pvrow_width : float, optional
            Width of the PV rows in the 2D plane in [m] (Default = None)
        param_names : list of str, optional
            List of surface parameter names for the PV surfaces
            (Default = None)
        cut : dict, optional
            Nested dictionary that tells if some PV row sides need to be
            discretized, and how (Default = None).
            Example: {1: {'front': 5}}, will create 5 segments on the front
            side of the PV row with index 1
        """
        # Initialize base parameters: common to all sorts of PV arrays
        super(OrderedPVArray, self).__init__(axis_azimuth=axis_azimuth)

        # These are the invariant parameters of the PV array
        self.gcr = gcr
        self.height = pvrow_height
        self.distance = (pvrow_width / gcr
                         if (pvrow_width is not None) and (gcr is not None)
                         else None)
        self.width = pvrow_width
        self.n_pvrows = n_pvrows
        self.param_names = [] if param_names is None else param_names
        self.cut = {} if cut is None else cut

        # These attributes will be updated at fitting time
        self.solar_2d_vectors = None
        self.n_states = None
        self.has_direct_shading = None
        self.rotation_vec = None
        self.shaded_length_front = None
        self.shaded_length_back = None

        # These attributes will be updated by the engine
        self.ts_vf_matrix = None
        self.ts_vf_aoi_matrix = None

    @classmethod
    def init_from_dict(cls, pvarray_params, param_names=None):
        """Instantiate ordered PV array from dictionary of parameters

        Parameters
        ----------
        pvarray_params : dict
            The parameters defining the PV array
        param_names : list of str, optional
            List of parameter names to pass to surfaces (Default = None)

        Returns
        -------
        OrderedPVArray
            Initialized Ordered PV Array
        """
        return cls(axis_azimuth=pvarray_params['axis_azimuth'],
                   gcr=pvarray_params['gcr'],
                   pvrow_height=pvarray_params['pvrow_height'],
                   n_pvrows=pvarray_params['n_pvrows'],
                   pvrow_width=pvarray_params['pvrow_width'],
                   cut=pvarray_params.get('cut', {}),
                   param_names=param_names)

    @classmethod
    def fit_from_dict_of_scalars(cls, pvarray_params, param_names=None):
        """Instantiate, and fit ordered PV array using dictionary
        of scalar inputs.

        Parameters
        ----------
        pvarray_params : dict
            The parameters used for instantiation, fitting, and transformation
        param_names : list of str, optional
            List of parameter names to pass to surfaces (Default = None)

        Returns
        -------
        OrderedPVArray
            Initialized, and fitted Ordered PV Array
        """

        # Create pv array
        pvarray = cls.init_from_dict(pvarray_params,
                                     param_names=param_names)

        # Fit pv array to scalar values
        solar_zenith = np.array([pvarray_params['solar_zenith']])
        solar_azimuth = np.array([pvarray_params['solar_azimuth']])
        surface_tilt = np.array([pvarray_params['surface_tilt']])
        surface_azimuth = np.array([pvarray_params['surface_azimuth']])
        pvarray.fit(solar_zenith, solar_azimuth,
                    surface_tilt, surface_azimuth)

        return pvarray

    def fit(self, solar_zenith, solar_azimuth, surface_tilt, surface_azimuth):
        """Fit the ordered PV array to the list of solar and surface angles.
        All intermediate PV array results necessary to build the geometries
        will be calculated here using vectorization as much as possible.

        Intemediate results include: PV row coordinates for all timestamps,
        ground element coordinates for all timestamps, cases of direct
        shading, ...

        Parameters
        ----------
        solar_zenith : array-like or float
            Solar zenith angles [deg]
        solar_azimuth : array-like or float
            Solar azimuth angles [deg]
        surface_tilt : array-like or float
            Surface tilt angles, from 0 to 180 [deg]
        surface_azimuth : array-like or float
            Surface azimuth angles [deg]
        """

        self.n_states = len(solar_zenith)
        # Calculate rotation angles
        rotation_vec = _get_rotation_from_tilt_azimuth(
            surface_azimuth, self.axis_azimuth, surface_tilt)
        # Save rotation vector
        self.rotation_vec = rotation_vec

        # Calculate the solar 2D vectors for all timestamps
        self.solar_2d_vectors = _get_solar_2d_vectors(
            solar_zenith, solar_azimuth, self.axis_azimuth)
        # Calculate the angle made by 2D sun vector and x-axis
        alpha_vec = np.arctan2(self.solar_2d_vectors[1],
                               self.solar_2d_vectors[0])

        # Calculate the coordinates of all PV rows for all timestamps
        self._calculate_pvrow_elements_coords(alpha_vec, rotation_vec)

        # Calculate ground elements coordinates for all timestamps
        self.ts_ground = TsGround.from_ts_pvrows_and_angles(
            self.ts_pvrows, alpha_vec, rotation_vec, y_ground=self.y_ground,
            flag_overlap=self.has_direct_shading,
            param_names=self.param_names)

        # Save surface rotation angles
        self.rotation_vec = rotation_vec

        # Index all timeseries surfaces
        self._index_all_ts_surfaces()

    def _calculate_pvrow_elements_coords(self, alpha_vec, rotation_vec):
        """Calculate PV row coordinate elements in a vectorized way, such as
        PV row boundary coordinates and shaded lengths.

        Parameters
        ----------
        alpha_vec : array-like or float
            Angle made by 2d solar vector and x-axis [rad]
        rotation_vec : array-like or float
            Rotation angle of the PV rows [deg]
        """
        # Initialize timeseries pv rows
        self.ts_pvrows = []

        # Calculate interrow direct shading lengths
        self._calculate_interrow_shading(alpha_vec, rotation_vec)

        # Calculate coordinates of segments of each pv row side
        xy_centers = [(X_ORIGIN_PVROWS + idx * self.distance,
                       self.height + self.y_ground)
                      for idx in range(self.n_pvrows)]
        tilted_to_left = rotation_vec > 0.
        for idx_pvrow, xy_center in enumerate(xy_centers):
            # A special treatment needs to be applied to shaded lengths for
            # the PV rows at the edge of the PV array
            if idx_pvrow == 0:
                # the leftmost row doesn't have left neighbors
                shaded_length_front = np.where(tilted_to_left, 0.,
                                               self.shaded_length_front)
                shaded_length_back = np.where(tilted_to_left,
                                              self.shaded_length_back, 0.)
            elif idx_pvrow == (self.n_pvrows - 1):
                # the rightmost row does have right neighbors
                shaded_length_front = np.where(tilted_to_left,
                                               self.shaded_length_front, 0.)
                shaded_length_back = np.where(tilted_to_left, 0.,
                                              self.shaded_length_back)
            else:
                # use calculated shaded lengths
                shaded_length_front = self.shaded_length_front
                shaded_length_back = self.shaded_length_back
            # Create timeseries PV rows and add it to the list
            self.ts_pvrows.append(TsPVRow.from_raw_inputs(
                xy_center, self.width, rotation_vec,
                self.cut.get(idx_pvrow, {}), shaded_length_front,
                shaded_length_back, index=idx_pvrow,
                param_names=self.param_names))

    def _calculate_interrow_shading(self, alpha_vec, rotation_vec):
        """Calculate the shaded length on front and back side of PV rows when
        direct shading happens, and in a vectorized way.

        Parameters
        ----------
        alpha_vec : array-like or float
            Angle made by 2d solar vector and x-axis [rad]
        rotation_vec : array-like or float
            Rotation angle of the PV rows [deg]
        """

        if self.n_pvrows > 1:
            # Calculate intermediate values for direct shading
            alpha_vec_deg = np.rad2deg(alpha_vec)
            theta_t = 90. - rotation_vec
            theta_t_rad = np.deg2rad(theta_t)
            beta = theta_t + alpha_vec_deg
            beta_rad = np.deg2rad(beta)
            delta = self.distance * (
                np.sin(theta_t_rad) - np.cos(theta_t_rad) * np.tan(beta_rad))
            # Calculate temporary shaded lengths
            tmp_shaded_length_front = np.maximum(0, self.width - delta)
            tmp_shaded_length_back = np.maximum(0, self.width + delta)
            # The shaded length can't be longer than PV row (meaning sun can't
            # be under the horizon or something...)
            self.shaded_length_front = np.where(
                tmp_shaded_length_front > self.width, 0,
                tmp_shaded_length_front)
            self.shaded_length_back = np.where(
                tmp_shaded_length_back > self.width, 0,
                tmp_shaded_length_back)
        else:
            # Since there's 1 row, there can't be any direct shading
            self.shaded_length_front = np.zeros(self.n_states)
            self.shaded_length_back = np.zeros(self.n_states)

        # Flag times when there's direct shading
        self.has_direct_shading = (
            (self.shaded_length_front > DISTANCE_TOLERANCE)
            | (self.shaded_length_back > DISTANCE_TOLERANCE))



"""Module with view factor calculation tools"""

from pvfactors.config import MIN_X_GROUND, MAX_X_GROUND, DISTANCE_TOLERANCE
from pvfactors.geometry.timeseries import TsLineCoords, TsPointCoords
from pvlib.tools import cosd, sind
import numpy as np


class VFTsMethods(object):
    """This class contains all the methods used to calculate timeseries
    view factors for all the surfaces in
    :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`"""

    def vf_pvrow_gnd_surf(self, ts_pvrows, ts_ground, tilted_to_left,
                          vf_matrix):
        """Calculate the view factors between timeseries PV row and ground
        surfaces, and assign it to the passed view factor matrix using
        the surface indices.

        Parameters
        ----------
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        ts_ground : :py:class:`~pvfactors.geometry.timeseries.TsGround`
            Timeseries ground of the PV array
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        vf_matrix : np.ndarray
            View factor matrix to update during calculation. Should have 3
            dimensions as follows: [n_surfaces, n_surfaces, n_timesteps]
        """

        n_pvrows = len(ts_pvrows)
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows):
            # Separate gnd surfaces depending on side
            left_gnd_surfaces = ts_ground.ts_surfaces_side_of_cut_point(
                'left', idx_pvrow)
            right_gnd_surfaces = ts_ground.ts_surfaces_side_of_cut_point(
                'right', idx_pvrow)
            # Front side
            front = ts_pvrow.front
            for pvrow_surf in front.all_ts_surfaces:
                if pvrow_surf.is_empty:
                    # do no run calculation for this surface
                    continue
                ts_length = pvrow_surf.length
                i = pvrow_surf.index
                for gnd_surf in left_gnd_surfaces:
                    if gnd_surf.is_empty:
                        # do no run this calculation
                        continue
                    j = gnd_surf.index
                    vf_pvrow_to_gnd, vf_gnd_to_pvrow = (
                        self.vf_pvrow_surf_to_gnd_surf_obstruction_hottel(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=False, is_left=True))
                    vf_matrix[i, j, :] = vf_pvrow_to_gnd
                    vf_matrix[j, i, :] = vf_gnd_to_pvrow
                for gnd_surf in right_gnd_surfaces:
                    if gnd_surf.is_empty:
                        # do no run this calculation
                        continue
                    j = gnd_surf.index
                    vf_pvrow_to_gnd, vf_gnd_to_pvrow = (
                        self.vf_pvrow_surf_to_gnd_surf_obstruction_hottel(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=False, is_left=False))
                    vf_matrix[i, j, :] = vf_pvrow_to_gnd
                    vf_matrix[j, i, :] = vf_gnd_to_pvrow
            # Back side
            back = ts_pvrow.back
            for pvrow_surf in back.all_ts_surfaces:
                if pvrow_surf.is_empty:
                    # do no run calculation for this surface
                    continue
                ts_length = pvrow_surf.length
                i = pvrow_surf.index
                for gnd_surf in left_gnd_surfaces:
                    if gnd_surf.is_empty:
                        # do no run this calculation
                        continue
                    j = gnd_surf.index
                    vf_pvrow_to_gnd, vf_gnd_to_pvrow = (
                        self.vf_pvrow_surf_to_gnd_surf_obstruction_hottel(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=True, is_left=True))
                    vf_matrix[i, j, :] = vf_pvrow_to_gnd
                    vf_matrix[j, i, :] = vf_gnd_to_pvrow
                for gnd_surf in right_gnd_surfaces:
                    if gnd_surf.is_empty:
                        # do no run this calculation
                        continue
                    j = gnd_surf.index
                    vf_pvrow_to_gnd, vf_gnd_to_pvrow = (
                        self.vf_pvrow_surf_to_gnd_surf_obstruction_hottel(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=True, is_left=False))
                    vf_matrix[i, j, :] = vf_pvrow_to_gnd
                    vf_matrix[j, i, :] = vf_gnd_to_pvrow

    def vf_pvrow_surf_to_gnd_surf_obstruction_hottel(
            self, pvrow_surf, pvrow_idx, n_pvrows, tilted_to_left,
            ts_pvrows, gnd_surf, pvrow_surf_length, is_back=True,
            is_left=True):
        """Calculate view factors from timeseries PV row surface to a
        timeseries ground surface. This will return the calculated view
        factors from the PV row surface to the ground surface, AND from the
        ground surface to the PV row surface (using reciprocity).

        Parameters
        ----------
        pvrow_surf : :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries PV row surface to use for calculation
        pvrow_idx : int
            Index of the timeseries PV row on the which the pvrow_surf is
        n_pvrows : int
            Number of timeseries PV rows in the PV array, and therefore number
            of shadows they cast on the ground
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        gnd_surf : :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries ground surface to use for calculation
        pvrow_surf_length : np.ndarray
            Length (width) of the timeseries PV row surface [m]
        is_back : bool
            Flag specifying whether pv row surface is on  back or front surface
            (Default = True)
        is_left : bool
            Flag specifying whether gnd surface is left of pv row cut point or
            not (Default = True)

        Returns
        -------
        vf_pvrow_to_gnd_surf : np.ndarray
            View factors from timeseries PV row surface to timeseries ground
            surface, dimension is [n_timesteps]
        vf_gnd_to_pvrow_surf : np.ndarray
            View factors from timeseries ground surface to timeseries PV row
            surface, dimension is [n_timesteps]
        """

        pvrow_surf_lowest_pt = pvrow_surf.lowest_point
        pvrow_surf_highest_pt = pvrow_surf.highest_point
        no_obstruction = (is_left & (pvrow_idx == 0)) \
            or ((not is_left) & (pvrow_idx == n_pvrows - 1))
        if no_obstruction:
            # There is no obstruction to the gnd surface
            vf_pvrow_to_gnd_surf = self._vf_surface_to_surface(
                pvrow_surf.coords, gnd_surf, pvrow_surf_length)
        else:
            # Get lowest point of obstructing point
            idx_obstructing_pvrow = pvrow_idx - 1 if is_left else pvrow_idx + 1
            pt_obstr = ts_pvrows[idx_obstructing_pvrow
                                 ].full_pvrow_coords.lowest_point
            # Calculate vf from pv row to gnd surface
            vf_pvrow_to_gnd_surf = self._vf_hottel_gnd_surf(
                pvrow_surf_highest_pt, pvrow_surf_lowest_pt,
                gnd_surf.b1, gnd_surf.b2, pt_obstr, pvrow_surf_length,
                is_left)

        # Final result depends on whether front or back surface
        if is_left:
            vf_pvrow_to_gnd_surf = (
                np.where(tilted_to_left, 0., vf_pvrow_to_gnd_surf) if is_back
                else np.where(tilted_to_left, vf_pvrow_to_gnd_surf, 0.))
        else:
            vf_pvrow_to_gnd_surf = (
                np.where(tilted_to_left, vf_pvrow_to_gnd_surf, 0.) if is_back
                else np.where(tilted_to_left, 0., vf_pvrow_to_gnd_surf))

        # Use reciprocity to calculate ts vf from gnd surf to pv row surface
        gnd_surf_length = gnd_surf.length
        vf_gnd_to_pvrow_surf = np.where(
            gnd_surf_length > DISTANCE_TOLERANCE,
            vf_pvrow_to_gnd_surf * pvrow_surf_length / gnd_surf_length, 0.)

        return vf_pvrow_to_gnd_surf, vf_gnd_to_pvrow_surf

    def vf_pvrow_to_pvrow(self, ts_pvrows, tilted_to_left, vf_matrix):
        """Calculate the view factors between timeseries PV row surfaces,
        and assign values to the passed view factor matrix using
        the surface indices.

        Parameters
        ----------
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        vf_matrix : np.ndarray
            View factor matrix to update during calculation. Should have 3
            dimensions as follows: [n_surfaces, n_surfaces, n_timesteps]
        """
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows[:-1]):
            # Get the next pv row
            right_ts_pvrow = ts_pvrows[idx_pvrow + 1]
            # front side
            front = ts_pvrow.front
            for surf_i in front.all_ts_surfaces:
                if surf_i.is_empty:
                    # do no run calculation for this surface
                    continue
                i = surf_i.index
                length_i = surf_i.length
                for surf_j in right_ts_pvrow.back.all_ts_surfaces:
                    if surf_j.is_empty:
                        # do no run calculation for this surface
                        continue
                    j = surf_j.index
                    length_j = surf_j.length
                    vf_i_to_j = self._vf_surface_to_surface(
                        surf_i.coords, surf_j.coords, length_i)
                    vf_i_to_j = np.where(tilted_to_left, 0., vf_i_to_j)
                    vf_j_to_i = np.where(
                        surf_j.length > DISTANCE_TOLERANCE,
                        vf_i_to_j * length_i / length_j, 0.)
                    vf_matrix[i, j, :] = vf_i_to_j
                    vf_matrix[j, i, :] = vf_j_to_i
            # back side
            back = ts_pvrow.back
            for surf_i in back.all_ts_surfaces:
                if surf_i.is_empty:
                    # do no run calculation for this surface
                    continue
                i = surf_i.index
                length_i = surf_i.length
                for surf_j in right_ts_pvrow.front.all_ts_surfaces:
                    if surf_j.is_empty:
                        # do no run calculation for this surface
                        continue
                    j = surf_j.index
                    length_j = surf_j.length
                    vf_i_to_j = self._vf_surface_to_surface(
                        surf_i.coords, surf_j.coords, length_i)
                    vf_i_to_j = np.where(tilted_to_left, vf_i_to_j, 0.)
                    vf_j_to_i = np.where(
                        surf_j.length > DISTANCE_TOLERANCE,
                        vf_i_to_j * length_i / length_j, 0.)
                    vf_matrix[i, j, :] = vf_i_to_j
                    vf_matrix[j, i, :] = vf_j_to_i

    def calculate_vf_to_pvrow(self, pvrow_element_coords, pvrow_idx, n_pvrows,
                              n_steps, ts_pvrows, pvrow_element_length,
                              tilted_to_left, pvrow_width, rotation_vec):
        """Calculate view factors from timeseries pvrow element to timeseries
        PV rows around it.

        Parameters
        ----------
        pvrow_element_coords :
        :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates of pvrow_element
        pvrow_idx : int
            Index of the timeseries PV row on which the pvrow_element is
        n_pvrows : int
            Number of timeseries PV rows in the PV array
        n_steps : int
            Number of timesteps for which to calculate the pvfactors
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            Timeseries PV row geometries that will be used in the calculation
        pvrow_element_length : float or np.ndarray
            Length (width) of the timeseries pvrow element [m]
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        pvrow_width : float
            Width of the timeseries PV rows in the PV array [m], which is
            constant
        rotation_vec : np.ndarray
            Rotation angles of the PV rows [deg]

        Returns
        -------
        vf_to_pvrow : np.ndarray
            View factors from timeseries pvrow_element to neighboring PV rows
        vf_to_shaded_pvrow : np.ndarray
            View factors from timeseries pvrow_element to shaded areas of the
            neighboring PV rows
        """
        if pvrow_idx == 0:
            vf_left_pvrow = np.zeros(n_steps)
            vf_left_shaded_pvrow = np.zeros(n_steps)
        else:
            # Get vf to full pvrow
            left_ts_pvrow = ts_pvrows[pvrow_idx - 1]
            left_ts_pvrow_coords = left_ts_pvrow.full_pvrow_coords
            vf_left_pvrow = self._vf_surface_to_surface(
                pvrow_element_coords, left_ts_pvrow_coords,
                pvrow_element_length)
            # Get vf to shaded pvrow
            shaded_coords = self._create_shaded_side_coords(
                left_ts_pvrow.xy_center, pvrow_width,
                left_ts_pvrow.front.shaded_length, tilted_to_left,
                rotation_vec, left_ts_pvrow.full_pvrow_coords.lowest_point)
            vf_left_shaded_pvrow = self._vf_surface_to_surface(
                pvrow_element_coords, shaded_coords, pvrow_element_length)

        if pvrow_idx == (n_pvrows - 1):
            vf_right_pvrow = np.zeros(n_steps)
            vf_right_shaded_pvrow = np.zeros(n_steps)
        else:
            # Get vf to full pvrow
            right_ts_pvrow = ts_pvrows[pvrow_idx + 1]
            right_ts_pvrow_coords = right_ts_pvrow.full_pvrow_coords
            vf_right_pvrow = self._vf_surface_to_surface(
                pvrow_element_coords, right_ts_pvrow_coords,
                pvrow_element_length)
            # Get vf to shaded pvrow
            shaded_coords = self._create_shaded_side_coords(
                right_ts_pvrow.xy_center, pvrow_width,
                right_ts_pvrow.front.shaded_length, tilted_to_left,
                rotation_vec, right_ts_pvrow.full_pvrow_coords.lowest_point)
            vf_right_shaded_pvrow = self._vf_surface_to_surface(
                pvrow_element_coords, shaded_coords, pvrow_element_length)

        vf_to_pvrow = np.where(tilted_to_left, vf_right_pvrow, vf_left_pvrow)
        vf_to_shaded_pvrow = np.where(tilted_to_left, vf_right_shaded_pvrow,
                                      vf_left_shaded_pvrow)

        return vf_to_pvrow, vf_to_shaded_pvrow

    def calculate_vf_to_gnd(self, pvrow_element_coords, pvrow_idx, n_pvrows,
                            n_steps, y_ground, cut_point_coords,
                            pvrow_element_length, tilted_to_left, ts_pvrows):
        """Calculate view factors from timeseries pvrow_element to the entire
        ground.

        Parameters
        ----------
        pvrow_element_coords :
        :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates of pvrow element
        pvrow_idx : int
            Index of the timeseries PV row on the which the pvrow_element is
        n_pvrows : int
            Number of timeseries PV rows in the PV array
        n_steps : int
            Number of timesteps for which to calculate the pvfactors
        y_ground : float
            Y-coordinate of the flat ground [m]
        cut_point_coords : list of
        :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            List of cut point coordinates, as calculated for timeseries PV rows
        pvrow_element_length : float or np.ndarray
            Length (width) of the timeseries pvrow_element [m]
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            Timeseries PV row geometries that will be used in the calculation

        Returns
        -------
        vf_to_gnd : np.ndarray
            View factors from timeseries pvrow_element to the entire ground
        """

        pvrow_lowest_pt = ts_pvrows[pvrow_idx].full_pvrow_coords.lowest_point
        if pvrow_idx == 0:
            # There is no obstruction to view of the ground on the left
            coords_left_gnd = TsLineCoords(
                TsPointCoords(MIN_X_GROUND * np.ones(n_steps), y_ground),
                TsPointCoords(np.minimum(MAX_X_GROUND, cut_point_coords.x),
                              y_ground))
            vf_left_ground = self._vf_surface_to_surface(
                pvrow_element_coords, coords_left_gnd, pvrow_element_length)
        else:
            # The left PV row obstructs the view of the ground on the left
            left_pt_neighbor = \
                ts_pvrows[pvrow_idx - 1].full_pvrow_coords.lowest_point
            coords_gnd_proxy = TsLineCoords(left_pt_neighbor, pvrow_lowest_pt)
            vf_left_ground = self._vf_surface_to_surface(
                pvrow_element_coords, coords_gnd_proxy, pvrow_element_length)

        if pvrow_idx == (n_pvrows - 1):
            # There is no obstruction of the view of the ground on the right
            coords_right_gnd = TsLineCoords(
                TsPointCoords(np.maximum(MIN_X_GROUND, cut_point_coords.x),
                              y_ground),
                TsPointCoords(MAX_X_GROUND * np.ones(n_steps), y_ground))
            vf_right_ground = self._vf_surface_to_surface(
                pvrow_element_coords, coords_right_gnd, pvrow_element_length)
        else:
            # The right PV row obstructs the view of the ground on the right
            right_pt_neighbor = \
                ts_pvrows[pvrow_idx + 1].full_pvrow_coords.lowest_point
            coords_gnd_proxy = TsLineCoords(pvrow_lowest_pt, right_pt_neighbor)
            vf_right_ground = self._vf_surface_to_surface(
                pvrow_element_coords, coords_gnd_proxy, pvrow_element_length)

        # Merge the views of the ground for the back side
        vf_ground = np.where(tilted_to_left, vf_right_ground, vf_left_ground)

        return vf_ground

    def calculate_vf_to_shadow_obstruction_hottel(
            self, pvrow_element, pvrow_idx, n_shadows, n_steps, tilted_to_left,
            ts_pvrows, shadow_left, shadow_right, pvrow_element_length):
        """Calculate view factors from timeseries pvrow_element to the shadow
        of a specific timeseries PV row which is casted on the ground.

        Parameters
        ----------
        pvrow_element : :py:class:`~pvfactors.geometry.timeseries.TsDualSegment` or :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries pvrow_element to use for calculation
        pvrow_idx : int
            Index of the timeseries PV row on the which the pvrow_element is
        n_shadows : int
            Number of timeseries PV rows in the PV array, and therefore number
            of shadows they cast on the ground
        n_steps : int
            Number of timesteps for which to calculate the pvfactors
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            Timeseries PV row geometries that will be used in the calculation
        shadow_left : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Coordinates of the shadow that are on the left side of the cut
            point of the PV row on which the pvrow_element is
        shadow_right : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Coordinates of the shadow that are on the right side of the cut
            point of the PV row on which the pvrow_element is
        pvrow_element_length : float or np.ndarray
            Length (width) of the timeseries pvrow_element [m]

        Returns
        -------
        vf_to_shadow : np.ndarray
            View factors from timeseries pvrow_element to the ground shadow of
            a specific timeseries PV row
        """

        pvrow_element_lowest_pt = pvrow_element.lowest_point
        pvrow_element_highest_pt = pvrow_element.highest_point
        # Calculate view factors to left shadows
        if pvrow_idx == 0:
            # There is no obstruction on the left
            vf_to_left_shadow = self._vf_surface_to_surface(
                pvrow_element.coords, shadow_left, pvrow_element_length)
        else:
            # There is potential obstruction on the left
            pt_obstr = ts_pvrows[pvrow_idx - 1].full_pvrow_coords.lowest_point
            is_shadow_left = True
            vf_to_left_shadow = self._vf_hottel_gnd_surf(
                pvrow_element_highest_pt, pvrow_element_lowest_pt,
                shadow_left.b1, shadow_left.b2, pt_obstr, pvrow_element_length,
                is_shadow_left)

        # Calculate view factors to right shadows
        if pvrow_idx == n_shadows - 1:
            # There is no obstruction on the right
            vf_to_right_shadow = self._vf_surface_to_surface(
                pvrow_element.coords, shadow_right, pvrow_element_length)
        else:
            # There is potential obstruction on the right
            pt_obstr = ts_pvrows[pvrow_idx + 1].full_pvrow_coords.lowest_point
            is_shadow_left = False
            vf_to_right_shadow = self._vf_hottel_gnd_surf(
                pvrow_element_highest_pt, pvrow_element_lowest_pt,
                shadow_right.b1, shadow_right.b2, pt_obstr,
                pvrow_element_length, is_shadow_left)

        # Filter since we're considering the back surface only
        vf_to_shadow = np.where(tilted_to_left, vf_to_right_shadow,
                                vf_to_left_shadow)

        return vf_to_shadow

    def _vf_hottel_gnd_surf(self, high_pt_pv, low_pt_pv, left_pt_gnd,
                            right_pt_gnd, obstr_pt, width, shadow_is_left):
        """
        Calculate the timeseries view factors from a PV surface defined by low
        and high points, to a ground surface defined by left and right points,
        while accounting for potentially obstructing neighboring PV rows,
        defined by an obstruction point, and all of this using the Hottel
        String method.

        Parameters
        ----------
        high_pt_pv : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Highest point of the PV surface, for each timestamp
        low_pt_pv : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Lowest point of the PV surface, for each timestamp
        left_pt_gnd : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Leftmost point  of the ground surface, for each timestamp
        right_pt_gnd : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Rightmost point  of the ground surface, for each timestamp
        obstr_pt : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Obstructing point of neighboring PV row, for each timestamp
        width : float or np.ndarray
            Width of the PV row surface considered, from low to high point [m]
        shadow_is_left : bool
            Side of the considered shadow (or ground) surface with respect to
            the edge point of the PV row on which the considered PV surface is
            located

        Returns
        -------
        vf_1_to_2 : np.ndarray
            View factors from PV surface to ground (shadow) surface
        """

        if shadow_is_left:
            # When the shadow is left
            # - uncrossed strings are high_pv - left_gnd and low_pv - right_gnd
            # - crossed strings are high_pv - right_gnd and low_pv - left_gnd
            l1 = self._hottel_string_length(high_pt_pv, left_pt_gnd, obstr_pt,
                                            shadow_is_left)
            l2 = self._hottel_string_length(low_pt_pv, right_pt_gnd, obstr_pt,
                                            shadow_is_left)
            d1 = self._hottel_string_length(high_pt_pv, right_pt_gnd, obstr_pt,
                                            shadow_is_left)
            d2 = self._hottel_string_length(low_pt_pv, left_pt_gnd, obstr_pt,
                                            shadow_is_left)
        else:
            # When the shadow is right
            # - uncrossed strings are high_pv - right_gnd and low_pv - left_gnd
            # - crossed strings are high_pv - left_gnd and low_pv - right_gnd
            l1 = self._hottel_string_length(high_pt_pv, right_pt_gnd, obstr_pt,
                                            shadow_is_left)
            l2 = self._hottel_string_length(low_pt_pv, left_pt_gnd, obstr_pt,
                                            shadow_is_left)
            d1 = self._hottel_string_length(high_pt_pv, left_pt_gnd, obstr_pt,
                                            shadow_is_left)
            d2 = self._hottel_string_length(low_pt_pv, right_pt_gnd, obstr_pt,
                                            shadow_is_left)
        vf_1_to_2 = (d1 + d2 - l1 - l2) / (2. * width)
        # The formula doesn't work if surface is a point
        vf_1_to_2 = np.where(width > DISTANCE_TOLERANCE, vf_1_to_2, 0.)

        return vf_1_to_2

    def _hottel_string_length(self, pt_pv, pt_gnd, pt_obstr, shadow_is_left):
        """
        Calculate a string length as defined by the Hottel String method in the
        calculation of view factors, which allows to account for obstructions.

        Parameters
        ----------
        left_pt_gnd : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Leftmost point  of the ground surface, for each timestamp
        right_pt_gnd : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Rightmost point  of the ground surface, for each timestamp
        obstr_pt : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Obstructing point of neighboring PV row, for each timestamp
        shadow_is_left : bool
            Side of the considered shadow (or ground) surface with respect to
            the edge point of the PV row on which the considered PV surface is
            located

        Returns
        -------
        hottel_length : np.ndarray
            Return timeseries length of the string, while accounting for
            obstructions, in [m]
        """
        # Calculate length of string without obstruction
        l_pv = self._distance(pt_pv, pt_gnd)
        if pt_obstr is None:
            # There can't be any obstruction
            hottel_length = l_pv
        else:
            # Determine if there is obstruction by using the angles made by
            # specific strings with the x-axis
            alpha_pv = self._angle_with_x_axis(pt_gnd, pt_pv)
            alpha_ob = self._angle_with_x_axis(pt_gnd, pt_obstr)
            if shadow_is_left:
                is_obstructing = alpha_pv > alpha_ob
            else:
                is_obstructing = alpha_pv < alpha_ob
            # Calculate length of string with obstruction
            l_obstr = (self._distance(pt_gnd, pt_obstr)
                       + self._distance(pt_obstr, pt_pv))
            # Merge based on whether there is obstruction or not
            hottel_length = np.where(is_obstructing, l_obstr, l_pv)
        return hottel_length

    def _vf_surface_to_surface(self, line_1, line_2, width_1):
        """Calculate view factors between timeseries line coords, and using
        the Hottel String method for calculating view factors (without
        obstruction).

        Parameters
        ----------
        line_1 : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates of surface 1
        line_2 : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates of surface 2
        width_1 : float or np.ndarray
            Length of line_1 in [m]

        Returns
        -------
        vf_1_to_2 : np.ndarray
            View factors from line_1 to line_2, for each timestep
        """
        length_1 = self._distance(line_1.b1, line_2.b1)
        length_2 = self._distance(line_1.b2, line_2.b2)
        length_3 = self._distance(line_1.b1, line_2.b2)
        length_4 = self._distance(line_1.b2, line_2.b1)
        sum_1 = length_1 + length_2
        sum_2 = length_3 + length_4
        vf_1_to_2 = np.abs(sum_2 - sum_1) / (2. * width_1)
        # The formula doesn't work if the line is a point
        vf_1_to_2 = np.where(width_1 > DISTANCE_TOLERANCE, vf_1_to_2, 0.)

        return vf_1_to_2

    @staticmethod
    def _distance(pt_1, pt_2):
        """Calculate distance between two timeseries points

        Parameters
        ----------
        pt_1 : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries point coordinates of point 1
        pt_2 : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries point coordinates of point 2

        Returns
        -------
        np.ndarray
            Distance between the two points, for each timestep
        """
        return np.sqrt((pt_2.y - pt_1.y)**2 + (pt_2.x - pt_1.x)**2)

    @staticmethod
    def _angle_with_x_axis(pt_1, pt_2):
        """Angle with x-axis of vector going from pt_1 to pt_2

        Parameters
        ----------
        pt_1 : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries point coordinates of point 1
        pt_2 : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries point coordinates of point 2

        Returns
        -------
        np.ndarray
            Angle between vector pt_1->pt_2 and x-axis
        """
        return np.arctan2(pt_2.y - pt_1.y, pt_2.x - pt_1.x)

    @staticmethod
    def _create_shaded_side_coords(xy_center, width, shaded_length,
                                   mask_tilted_to_left, rotation_vec,
                                   side_lowest_pt):
        """
        Create the timeseries line coordinates for the shaded portion of a
        PV row side, based on inputted shaded length.

        Parameters
        ----------
        xy_center : tuple of float
            x and y coordinates of the PV row center point (invariant)
        width : float
            width of the PV rows [m]
        shaded_length : np.ndarray
            Timeseries values of side shaded length [m]
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        rotation_vec : np.ndarray
            Timeseries rotation vector of the PV rows in [deg]
        side_lowest_pt : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of lowest point of considered PV row side

        Returns
        -------
        side_shaded_coords : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of the shaded portion of the PV row side
        """

        # Get invariant values
        x_center, y_center = xy_center
        radius = width / 2.

        # Calculate coords of shading point
        r_shade = radius - shaded_length
        x_sh = np.where(
            mask_tilted_to_left,
            r_shade * cosd(rotation_vec + 180.) + x_center,
            r_shade * cosd(rotation_vec) + x_center)
        y_sh = np.where(
            mask_tilted_to_left,
            r_shade * sind(rotation_vec + 180.) + y_center,
            r_shade * sind(rotation_vec) + y_center)

        side_shaded_coords = TsLineCoords(TsPointCoords(x_sh, y_sh),
                                          side_lowest_pt)

        return side_shaded_coords



"""Module with Base classes for irradiance models"""
import numpy as np


class BaseModel(object):
    """Base class for irradiance models"""

    params = None
    cats = None
    irradiance_comp = None

    def fit(self, *args, **kwargs):
        """Not implemented"""
        raise NotImplementedError

    def transform(self, *args, **kwargs):
        """Not implemented"""
        raise NotImplementedError

    def get_full_modeling_vectors(self, *args, **kwargs):
        """Not implemented"""
        raise NotImplementedError

    @property
    def gnd_shaded(self):
        """Not implemented"""
        raise NotImplementedError

    @property
    def gnd_illum(self):
        """Not implemented"""
        raise NotImplementedError

    @property
    def pvrow_shaded(self):
        """Not implemented"""
        raise NotImplementedError

    @property
    def pvrow_illum(self):
        """Not implemented"""
        raise NotImplementedError

    @property
    def sky_luminance(self):
        """Not implemented"""
        raise NotImplementedError

    def get_ts_modeling_vectors(self, pvarray):
        """Get matrices of summed up irradiance values from a PV array, as well
        as the inverse reflectivity values (the latter need to be named
        "inv_rho"), and the total perez irradiance values.

        Parameters
        ----------
        pvarray : PV array object
            PV array with the irradiance and reflectivity values

        Returns
        -------
        irradiance_mat : list
            Matrix of summed up non-reflective irradiance values for all
            timeseries surfaces. Dimension = [n_surfaces, n_timesteps]
        rho_mat : list
            Matrix of reflectivity values for all timeseries surfaces
            Dimension = [n_surfaces, n_timesteps]
        invrho_mat : list
            List of inverse reflectivity for all timeseries surfaces
            Dimension = [n_surfaces, n_timesteps]
        total_perez_mat : list
            List of total perez transposed irradiance values for all timeseries
            surfaces
            Dimension = [n_surfaces, n_timesteps]
        """

        # TODO: this can probably be speeded up
        irradiance_mat = []
        rho_mat = []
        invrho_mat = []
        total_perez_mat = []
        # In principle, the list all ts surfaces should be ordered
        # with the ts surfaces' indices. ie first element has index 0, etc
        for ts_surface in pvarray.all_ts_surfaces:
            value = np.zeros(pvarray.n_states, dtype=float)
            for component in self.irradiance_comp:
                value += ts_surface.get_param(component)
            irradiance_mat.append(value)
            invrho_mat.append(ts_surface.get_param('inv_rho'))
            rho_mat.append(ts_surface.get_param('rho'))
            total_perez_mat.append(ts_surface.get_param('total_perez'))

        return irradiance_mat, rho_mat, invrho_mat, total_perez_mat

    def get_summed_components(self, pvarray, absorbed=True):
        """Get sum of irradiance components for irradiance model,
        either absorbed or only incident.

        Parameters
        ----------
        pvarray : PV array object
            PV array with the irradiance and reflectivity values
        absorbed : bool, optional
            Flag to decide whether to use the absorbed components are not
            (default = True)

        Returns
        -------
        irradiance_mat : list
            Matrix of summed up non-reflective irradiance values for all
            timeseries surfaces. Dimension = [n_surfaces, n_timesteps]"""
        # Initialize
        list_components = (self.irradiance_comp_absorbed if absorbed
                           else self.irradiance_comp)
        # Build list
        irradiance_mat = []
        for ts_surface in pvarray.all_ts_surfaces:
            value = np.zeros(pvarray.n_states, dtype=float)
            for component in list_components:
                value += ts_surface.get_param(component)
            irradiance_mat.append(value)
        return np.array(irradiance_mat)

    def update_ts_surface_sky_term(self, ts_surface, name_sky_term='sky_term'):
        """Update the 'sky_term' parameter of a timeseries surface.

        Parameters
        ----------
        ts_surface :  :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries surface whose 'sky_term' parameter value we want to
            update
        name_sky_term : str, optional
            Name of the sky term parameter (Default = 'sky_term')
        """
        value = 0.
        for component in self.irradiance_comp:
            value += ts_surface.get_param(component)
        ts_surface.update_params({name_sky_term: value})

    def initialize_rho(self, rho_scalar, rho_calculated, default_value):
        """Initialize reflectivity value:
        - if a scalar value is passed, use it
        - otherwise try to use calculated value
        - else use default value

        Parameters
        ----------
        rho_scalar : float
            Global average reflectivity value that is supposed to be used
        rho_calculated : float
            Reflectivity value calculated
        default_value : float
            Default value to use if everything fails

        Returns
        -------
        rho_scalar : float
            Global average reflectivity
        """
        if np.isscalar(rho_scalar):
            rho_scalar = rho_scalar
        elif np.isscalar(rho_calculated):
            rho_scalar = rho_calculated
        else:
            rho_scalar = default_value
        return rho_scalar



"""Utility functions for geometrical calculations."""

import numpy as np
from pvfactors import PVFactorsError
from pvfactors.config import TOL_COLLINEAR, DISTANCE_TOLERANCE
from shapely.geometry import \
    Point, GeometryCollection, LineString, MultiLineString


def difference(u, v):
    """Calculate difference between two lines, avoiding shapely float
    precision errors

    Parameters
    ----------
    u : :py:class:`shapely.geometry.LineString`-like
        Line string from which ``v`` will be removed
    v : :py:class:`shapely.geometry.LineString`-like
        Line string to remove from ``u``

    Returns
    -------
    :py:class:`shapely.geometry.LineString`
       Resulting difference of current surface minus given linestring
    """
    ub1, ub2 = u.boundary
    vb1, vb2 = v.boundary
    u_contains_vb1 = contains(u, vb1)
    u_contains_vb2 = contains(u, vb2)
    v_contains_ub1 = contains(v, ub1)
    v_contains_ub2 = contains(v, ub2)

    if u_contains_vb1:
        if u_contains_vb2:
            l_tmp = LineString([ub1, vb1])
            if contains(l_tmp, vb2):
                list_lines = [LineString([ub1, vb2]), LineString([vb1, ub2])]
            else:
                list_lines = [LineString([ub1, vb1]), LineString([vb2, ub2])]
            # Note that boundary points can be equal, so need to make sure
            # we're not passing line strings with length 0
            final_list_lines = [line for line in list_lines
                                if line.length > DISTANCE_TOLERANCE]
            len_final_list = len(final_list_lines)
            if len_final_list == 2:
                return MultiLineString(final_list_lines)
            elif len_final_list == 1:
                return final_list_lines[0]
            else:
                return LineString()
        elif v_contains_ub1:
            if v_contains_ub2:
                return LineString()
            else:
                return LineString([vb1, ub2])
        elif v_contains_ub2:
            return LineString([ub1, vb1])
        else:
            return u
    elif u_contains_vb2:
        if v_contains_ub1:
            if v_contains_ub2:
                return LineString()
            else:
                return LineString([vb2, ub2])
        elif v_contains_ub2:
            return LineString([ub1, vb2])
        else:
            return u
    else:
        return u


def contains(linestring, point, tol_distance=DISTANCE_TOLERANCE):
    """Fixing floating point errors obtained in shapely for contains"""
    return linestring.distance(point) < tol_distance


def is_collinear(list_elements):
    """Check that all :py:class:`~pvfactors.pvsurface.PVSegment`
    or :py:class:`~pvfactors.pvsurface.PVSurface` objects in list
    are collinear"""
    is_col = True
    u_direction = None  # will be orthogonal to normal vector
    for element in list_elements:
        if not element.is_empty:
            if u_direction is None:
                # set u_direction if not defined already
                u_direction = np.array([- element.n_vector[1],
                                        element.n_vector[0]])
            else:
                # check that collinear
                dot_prod = u_direction.dot(element.n_vector)
                is_col = np.abs(dot_prod) < TOL_COLLINEAR
                if not is_col:
                    return is_col
    return is_col


def check_collinear(list_elements):
    """Raise error if all :py:class:`~pvfactors.pvsurface.PVSegment`
    or :py:class:`~pvfactors.pvsurface.PVSurface` objects in list
    are not collinear"""
    is_col = is_collinear(list_elements)
    if not is_col:
        msg = "All elements should be collinear"
        raise PVFactorsError(msg)


def are_2d_vecs_collinear(u1, u2):
    """Check that two 2D vectors are collinear"""
    n1 = np.array([-u1[1], u1[0]])
    dot_prod = n1.dot(u2)
    return np.abs(dot_prod) < TOL_COLLINEAR


def projection(point, vector, linestring, must_contain=True):
    """Projection of point along vector onto a linestring.
    Define equations of two lines:
    - one defined by point and vector: a*x + b*y + c = 0
    - one defined by linestring: d*x + e*y + f = 0

    then if the lines are not parallel, the interesction is defined by:
    X = - W^(-1) . B, where X = [x, y], W = [[a, b], [c, d]], B = [c, f]

    Do not use if the two lines are parallel (determinant is 0, W not
    invertible)
    """
    # Define equation a*x + b*y + c = 0
    a, b = -vector[1], vector[0]
    c = - (a * point.x + b * point.y)
    # Define equation d*x + e*y +f = 0
    b1, b2 = linestring.boundary
    d, e = - (b2.y - b1.y), b2.x - b1.x
    f = - (d * b1.x + e * b1.y)
    # TODO: check that two lines are not parallel
    n1 = [a, b]
    n2 = [d, e]
    if are_2d_vecs_collinear(n1, n2):
        return GeometryCollection()
    else:
        W = [[a, b], [d, e]]
        B = [c, f]
        x, y = - np.linalg.inv(W).dot(B)
        pt_intersection = Point(x, y)
        length_linestring = linestring.length
        # For the following, using linestring.contains(pt_intersection) leads
        # to wrong assessments sometimes, probably because of round off errors
        distance_to_b1 = b1.distance(pt_intersection)
        distance_to_b2 = b2.distance(pt_intersection)
        contained_by_linestring = (
            (linestring.distance(pt_intersection) < DISTANCE_TOLERANCE) and
            (distance_to_b1 <= length_linestring) and
            (distance_to_b2 <= length_linestring))
        if not must_contain:
            # No need for the geometry to contain it
            return pt_intersection
        elif contained_by_linestring:
            # Check that the intersection is not too close to a boundary: if it
            # is it can create a "memory access error" it seems
            too_close_to_b1 = distance_to_b1 < DISTANCE_TOLERANCE
            too_close_to_b2 = distance_to_b2 < DISTANCE_TOLERANCE
            if too_close_to_b1:
                return b1
            elif too_close_to_b2:
                return b2
            else:
                return pt_intersection
        else:
            return GeometryCollection()



"""Module containing AOI loss calculation methods"""

from pvfactors.config import DISTANCE_TOLERANCE
from pvfactors.geometry.timeseries import (
    TsPointCoords, TsSurface, TsLineCoords)
from pvfactors import PVFactorsError
import pvlib
from pvlib.tools import cosd
import numpy as np


class AOIMethods:
    """Class containing methods related to calculating AOI losses for
    :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray` objects."""

    def __init__(self, faoi_fn_front, faoi_fn_back, n_integral_sections=300):
        """Instantiate class with faoi function and number of sections to use
        to calculate integrals of view factors with faoi losses

        Parameters
        ----------
        faoi_fn_front : function
            Function which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the front side of PV rows
        faoi_fn_back : function
            Function which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the back side of PV rows
        n_integral_sections : int, optional
            Number of integral divisions of the 0 to 180 deg interval
            to use for the fAOI loss integral (default = 300)
        """
        # Check that faoi fn where passed
        faoi_fns_ok = callable(faoi_fn_front) and callable(faoi_fn_back)
        if not faoi_fns_ok:
            raise PVFactorsError("The faoi_fn passed to the AOI methods are "
                                 "not callable. Please check the fAOI "
                                 "functions again")
        self.faoi_fn_front = faoi_fn_front
        self.faoi_fn_back = faoi_fn_back
        self.n_integral_sections = n_integral_sections
        # The following will be updated at fitting time
        self.interval = None
        self.aoi_angles_low = None
        self.aoi_angles_high = None
        self.integrand_front = None
        self.integrand_back = None

    def fit(self, n_timestamps):
        """Fit the AOI methods to timeseries inputs: create all the necessary
        integration attributes.

        Parameters
        ----------
        n_timestamps : int
            Number of simulation timestamps
        """
        # Will use x values at the middle of the integral sections
        aoi_angles = np.linspace(0., 180., num=self.n_integral_sections + 1)
        # Assumes that at least 2 aoi angle values, otherwise what's the point
        self.interval = aoi_angles[1] - aoi_angles[0]
        # Get integral intervals' low, high, and middle points
        aoi_angles_low = aoi_angles[:-1]
        aoi_angles_high = aoi_angles_low + self.interval
        aoi_angles_middle = aoi_angles_low + self.interval / 2.
        # Calculate faoi values using middle points of integral intervals
        faoi_front = self.faoi_fn_front(aoi_angles_middle)
        faoi_back = self.faoi_fn_back(aoi_angles_middle)
        # Calculate small view factor values for each section
        vf_values = self._vf(aoi_angles_low, aoi_angles_high)
        # Multiply to get integrand
        integrand_front = faoi_front * vf_values
        integrand_back = faoi_back * vf_values
        # Replicate these values for all timestamps such that shapes
        # becomes: [n_timestamps, n_integral_sections]map
        self.aoi_angles_low = np.tile(aoi_angles_low, (n_timestamps, 1))
        self.aoi_angles_high = np.tile(aoi_angles_high, (n_timestamps, 1))
        self.integrand_front = np.tile(integrand_front, (n_timestamps, 1))
        self.integrand_back = np.tile(integrand_back, (n_timestamps, 1))

    def vf_aoi_pvrow_to_sky(self, ts_pvrows, ts_ground, tilted_to_left,
                            vf_matrix):
        """Calculate the view factors between timeseries PV row surface and sky
        while accounting for AOI losses,
        and assign values to the passed view factor matrix using
        the surface indices.

        Parameters
        ----------
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        ts_ground : :py:class:`~pvfactors.geometry.timeseries.TsGround`
            Timeseries ground of the PV array
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        vf_matrix : np.ndarray
            View factor matrix to update during calculation. Should have 3
            dimensions as follows: [n_surfaces, n_surfaces, n_timesteps]
        """
        sky_index = vf_matrix.shape[0] - 1
        # --- Build list of dummy sky surfaces
        # create sky left open area
        pt_1 = TsPointCoords(ts_ground.x_min * np.ones_like(tilted_to_left),
                             ts_ground.y_ground * np.ones_like(tilted_to_left))
        pt_2 = ts_pvrows[0].highest_point
        sky_left = TsSurface(TsLineCoords(pt_1, pt_2))
        # create sky right open area
        pt_1 = TsPointCoords(ts_ground.x_max * np.ones_like(tilted_to_left),
                             ts_ground.y_ground * np.ones_like(tilted_to_left))
        pt_2 = ts_pvrows[-1].highest_point
        sky_right = TsSurface(TsLineCoords(pt_2, pt_1))
        # Add sky surfaces in-between PV rows
        dummy_sky_surfaces = [sky_left]
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows[:-1]):
            right_ts_pvrow = ts_pvrows[idx_pvrow + 1]
            pt_1 = ts_pvrow.highest_point
            pt_2 = right_ts_pvrow.highest_point
            sky_surface = TsSurface(TsLineCoords(pt_1, pt_2))
            dummy_sky_surfaces.append(sky_surface)
        # Add sky right open area
        dummy_sky_surfaces.append(sky_right)

        # Now calculate vf_aoi for all PV row surfaces to sky
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows):
            # Get dummy sky surfaces
            sky_left = dummy_sky_surfaces[idx_pvrow]
            sky_right = dummy_sky_surfaces[idx_pvrow + 1]
            # Calculate vf_aoi for surfaces in PV row
            # front side
            front = ts_pvrow.front
            for front_surf in front.all_ts_surfaces:
                vf_aoi_left = self._vf_aoi_surface_to_surface(
                    front_surf, sky_left, is_back=False)
                vf_aoi_right = self._vf_aoi_surface_to_surface(
                    front_surf, sky_right, is_back=False)
                vf_aoi = np.where(tilted_to_left, vf_aoi_left, vf_aoi_right)
                vf_matrix[front_surf.index, sky_index, :] = vf_aoi
            # back side
            back = ts_pvrow.back
            for back_surf in back.all_ts_surfaces:
                vf_aoi_left = self._vf_aoi_surface_to_surface(
                    back_surf, sky_left, is_back=True)
                vf_aoi_right = self._vf_aoi_surface_to_surface(
                    back_surf, sky_right, is_back=True)
                vf_aoi = np.where(tilted_to_left, vf_aoi_right, vf_aoi_left)
                vf_matrix[back_surf.index, sky_index, :] = vf_aoi

    def vf_aoi_pvrow_to_pvrow(self, ts_pvrows, tilted_to_left, vf_matrix):
        """Calculate the view factors between timeseries PV row surfaces
        while accounting for AOI losses,
        and assign values to the passed view factor matrix using
        the surface indices.

        Parameters
        ----------
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        vf_matrix : np.ndarray
            View factor matrix to update during calculation. Should have 3
            dimensions as follows: [n_surfaces, n_surfaces, n_timesteps]
        """
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows[:-1]):
            # Get the next pv row
            right_ts_pvrow = ts_pvrows[idx_pvrow + 1]
            # front side
            front = ts_pvrow.front
            for surf_i in front.all_ts_surfaces:
                i = surf_i.index
                for surf_j in right_ts_pvrow.back.all_ts_surfaces:
                    j = surf_j.index
                    # vf aoi from i to j
                    vf_i_to_j = self._vf_aoi_surface_to_surface(
                        surf_i, surf_j, is_back=False)
                    vf_i_to_j = np.where(tilted_to_left, 0., vf_i_to_j)
                    # vf aoi from j to i
                    vf_j_to_i = self._vf_aoi_surface_to_surface(
                        surf_j, surf_i, is_back=True)
                    vf_j_to_i = np.where(tilted_to_left, 0., vf_j_to_i)
                    # save results
                    vf_matrix[i, j, :] = vf_i_to_j
                    vf_matrix[j, i, :] = vf_j_to_i
            # back side
            back = ts_pvrow.back
            for surf_i in back.all_ts_surfaces:
                i = surf_i.index
                for surf_j in right_ts_pvrow.front.all_ts_surfaces:
                    j = surf_j.index
                    # vf aoi from i to j
                    vf_i_to_j = self._vf_aoi_surface_to_surface(
                        surf_i, surf_j, is_back=True)
                    vf_i_to_j = np.where(tilted_to_left, vf_i_to_j, 0.)
                    # vf aoi from j to i
                    vf_j_to_i = self._vf_aoi_surface_to_surface(
                        surf_j, surf_i, is_back=False)
                    vf_j_to_i = np.where(tilted_to_left, vf_j_to_i, 0.)
                    # save results
                    vf_matrix[i, j, :] = vf_i_to_j
                    vf_matrix[j, i, :] = vf_j_to_i

    def vf_aoi_pvrow_to_gnd(self, ts_pvrows, ts_ground, tilted_to_left,
                            vf_aoi_matrix):
        """Calculate the view factors between timeseries PV row and ground
        surfaces while accounting for non-diffuse AOI losses,
        and assign it to the passed view factor aoi matrix using
        the surface indices.

        Notes
        -----
        This assumes that the PV row surfaces are infinitesimal (very small)

        Parameters
        ----------
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        ts_ground : :py:class:`~pvfactors.geometry.timeseries.TsGround`
            Timeseries ground of the PV array
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        vf_aoi_matrix : np.ndarray
            View factor aoi matrix to update during calculation. Should have 3
            dimensions as follows: [n_surfaces, n_surfaces, n_timesteps]
        """
        n_pvrows = len(ts_pvrows)
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows):
            # Separate gnd surfaces depending on side
            left_gnd_surfaces = ts_ground.ts_surfaces_side_of_cut_point(
                'left', idx_pvrow)
            right_gnd_surfaces = ts_ground.ts_surfaces_side_of_cut_point(
                'right', idx_pvrow)
            # Front side
            front = ts_pvrow.front
            for pvrow_surf in front.all_ts_surfaces:
                ts_length = pvrow_surf.length
                i = pvrow_surf.index
                for gnd_surf in left_gnd_surfaces:
                    j = gnd_surf.index
                    vf_pvrow_to_gnd = (
                        self._vf_aoi_pvrow_surf_to_gnd_surf_obstruction(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=False, is_left=True))
                    vf_aoi_matrix[i, j, :] = vf_pvrow_to_gnd
                for gnd_surf in right_gnd_surfaces:
                    j = gnd_surf.index
                    vf_pvrow_to_gnd = (
                        self._vf_aoi_pvrow_surf_to_gnd_surf_obstruction(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=False, is_left=False))
                    vf_aoi_matrix[i, j, :] = vf_pvrow_to_gnd
            # Back side
            back = ts_pvrow.back
            for pvrow_surf in back.all_ts_surfaces:
                ts_length = pvrow_surf.length
                i = pvrow_surf.index
                for gnd_surf in left_gnd_surfaces:
                    j = gnd_surf.index
                    vf_pvrow_to_gnd = (
                        self._vf_aoi_pvrow_surf_to_gnd_surf_obstruction(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=True, is_left=True))
                    vf_aoi_matrix[i, j, :] = vf_pvrow_to_gnd
                for gnd_surf in right_gnd_surfaces:
                    j = gnd_surf.index
                    vf_pvrow_to_gnd = (
                        self._vf_aoi_pvrow_surf_to_gnd_surf_obstruction(
                            pvrow_surf, idx_pvrow, n_pvrows,
                            tilted_to_left, ts_pvrows, gnd_surf, ts_length,
                            is_back=True, is_left=False))
                    vf_aoi_matrix[i, j, :] = vf_pvrow_to_gnd

    def _vf_aoi_surface_to_surface(self, surf_1, surf_2, is_back=True):
        """Calculate view factor, while accounting from AOI losses, from
        surface 1 to surface 2.

        Notes
        -----
        This assumes that surf_1 is infinitesimal (very small)

        Parameters
        ----------
        surf_1 : :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Infinitesimal surface from which to calculate view factor with
            AOI losses
        surf_2 : :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Surface to which the view factor with AOI losses should be
            calculated
        is_back : bool
            Flag specifying whether pv row surface is on back or front side
            of PV row (Default = True)

        Returns
        -------
        vf_aoi : np.ndarray
            View factors with aoi losses from surface 1 to surface 2,
            dimension is [n_timesteps]
        """
        # skip calculation if either surface is empty (always zero length)
        skip = surf_1.is_empty or surf_2.is_empty
        if skip:
            vf_aoi = np.zeros_like(surf_2.length)
        else:
            # Get surface 1 params
            u_vector = surf_1.u_vector
            centroid = surf_1.centroid
            # Calculate AOI angles
            aoi_angles_1 = self._calculate_aoi_angles(u_vector, centroid,
                                                      surf_2.b1)
            aoi_angles_2 = self._calculate_aoi_angles(u_vector, centroid,
                                                      surf_2.b2)
            low_aoi_angles = np.where(aoi_angles_1 < aoi_angles_2, aoi_angles_1,
                                      aoi_angles_2)
            high_aoi_angles = np.where(aoi_angles_1 < aoi_angles_2, aoi_angles_2,
                                       aoi_angles_1)
            # Calculate vf_aoi
            vf_aoi_raw = self._calculate_vf_aoi_wedge_level(
                low_aoi_angles, high_aoi_angles, is_back=is_back)
            # Should be zero where either of the surfaces have zero length
            vf_aoi = np.where((surf_1.length < DISTANCE_TOLERANCE)
                              | (surf_2.length < DISTANCE_TOLERANCE), 0.,
                              vf_aoi_raw)
        return vf_aoi

    def _vf_aoi_pvrow_surf_to_gnd_surf_obstruction(
            self, pvrow_surf, pvrow_idx, n_pvrows, tilted_to_left, ts_pvrows,
            gnd_surf, ts_length, is_back=True, is_left=True):
        """Calculate view factors from timeseries PV row surface to a
        timeseries ground surface, accounting for AOI losses.
        This will return the calculated view
        factors from the PV row surface to the ground surface.

        Notes
        -----
        This assumes that the PV row surfaces are infinitesimal (very small)

        Parameters
        ----------
        pvrow_surf : :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries PV row surface to use for calculation
        pvrow_idx : int
            Index of the timeseries PV row on the which the pvrow_surf is
        n_pvrows : int
            Number of timeseries PV rows in the PV array, and therefore number
            of shadows they cast on the ground
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        gnd_surf : :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries ground surface to use for calculation
        pvrow_surf_length : np.ndarray
            Length (width) of the timeseries PV row surface [m]
        is_back : bool
            Flag specifying whether pv row surface is on back or front side
            of PV row (Default = True)
        is_left : bool
            Flag specifying whether gnd surface is left of pv row cut point or
            not (Default = True)

        Returns
        -------
        vf_aoi_pvrow_to_gnd_surf : np.ndarray
            View factors aoi from timeseries PV row surface to timeseries
            ground surface, dimension is [n_timesteps]
        """
        # skip calculation if either surface is empty (always zero length)
        skip = pvrow_surf.is_empty or gnd_surf.is_empty
        if skip:
            vf_aoi = np.zeros_like(gnd_surf.length)
        else:
            centroid = pvrow_surf.centroid
            u_vector = pvrow_surf.u_vector
            no_obstruction = (is_left & (pvrow_idx == 0)) \
                or ((not is_left) & (pvrow_idx == n_pvrows - 1))
            if no_obstruction:
                # There is no obstruction to the ground surface
                aoi_angles_1 = self._calculate_aoi_angles(u_vector, centroid,
                                                          gnd_surf.b1)
                aoi_angles_2 = self._calculate_aoi_angles(u_vector, centroid,
                                                          gnd_surf.b2)
            else:
                # Get lowest point of obstructing point
                idx_obstructing_pvrow = (pvrow_idx - 1 if is_left
                                         else pvrow_idx + 1)
                pt_obstr = ts_pvrows[idx_obstructing_pvrow
                                     ].full_pvrow_coords.lowest_point
                # Adjust angle seen when there is obstruction
                aoi_angles_1 = self._calculate_aoi_angles_w_obstruction(
                    u_vector, centroid, gnd_surf.b1, pt_obstr, is_left)
                aoi_angles_2 = self._calculate_aoi_angles_w_obstruction(
                    u_vector, centroid, gnd_surf.b2, pt_obstr, is_left)

            low_aoi_angles = np.where(aoi_angles_1 < aoi_angles_2,
                                      aoi_angles_1, aoi_angles_2)
            high_aoi_angles = np.where(aoi_angles_1 < aoi_angles_2,
                                       aoi_angles_2, aoi_angles_1)
            vf_aoi_raw = self._calculate_vf_aoi_wedge_level(
                low_aoi_angles, high_aoi_angles, is_back=is_back)
            # Should be zero where either of the surfaces have zero length
            vf_aoi_raw = np.where((ts_length < DISTANCE_TOLERANCE)
                                  | (gnd_surf.length < DISTANCE_TOLERANCE), 0.,
                                  vf_aoi_raw)

            # Final result depends on whether front or back surface
            if is_left:
                vf_aoi = (np.where(tilted_to_left, 0., vf_aoi_raw) if is_back
                          else np.where(tilted_to_left, vf_aoi_raw, 0.))
            else:
                vf_aoi = (np.where(tilted_to_left, vf_aoi_raw, 0.) if is_back
                          else np.where(tilted_to_left, 0., vf_aoi_raw))

        return vf_aoi

    def _calculate_aoi_angles_w_obstruction(
            self, u_vector, centroid, point_gnd, point_obstr,
            gnd_surf_is_left):
        """Calculate AOI angles for a PV row surface of the
        :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray` that sees
        a ground surface, while being potentially obstructed by another
        PV row

        Parameters
        ----------
        u_vector : np.ndarray
            Direction vector of the surface for which to calculate AOI angles
        centroid : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Centroid point of PV row surface for which to calculate AOI angles
        point : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Point of ground surface that will determine AOI angle
        point_obstr: :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Potentially obstructing point for the view aoi angle calculation
        gnd_surf_is_left : bool
            Flag specifying whether ground surface is left of PV row's cut
            point or not

        Returns
        -------
        np.ndarray
            AOI angles formed by remote point and centroid on surface,
            measured against surface direction vector, accounting for
            potential obstruction [degrees]
        """
        if point_obstr is None:
            # There is no obstruction
            point = point_gnd
        else:
            # Determine if there is obstruction by using the angles made by
            # specific strings with the x-axis
            alpha_pv = self._angle_with_x_axis(point_gnd, centroid)
            alpha_ob = self._angle_with_x_axis(point_gnd, point_obstr)
            if gnd_surf_is_left:
                is_obstructing = alpha_pv > alpha_ob
            else:
                is_obstructing = alpha_pv < alpha_ob
            x = np.where(is_obstructing, point_obstr.x, point_gnd.x)
            y = np.where(is_obstructing, point_obstr.y, point_gnd.y)
            point = TsPointCoords(x, y)

        aoi_angles = self._calculate_aoi_angles(u_vector, centroid, point)
        return aoi_angles

    def _calculate_vf_aoi_wedge_level(self, low_angles, high_angles,
                                      is_back=True):
        """Calculate faoi modified view factors for a wedge defined by
        low and high angles.

        Parameters
        ----------
        low_angles : np.ndarray
            Low AOI angles (between 0 and 180 deg), length = n_timestamps
        high_angles : np.ndarray
            High AOI angles (between 0 and 180 deg), length = n_timestamps.
            Should be bigger than ``low_angles``
        is_back : bool
            Flag specifying whether pv row surface is on back or front side
            of PV row (Default = True)

        Returns
        -------
        np.ndarray
            faoi modified view factors for wedge
            shape = (n_timestamps, )
        """
        # Calculate integrand: all d_vf_aoi values
        faoi_integrand = self._calculate_vfaoi_integrand(
            low_angles, high_angles, is_back=is_back)
        # Total vf_aoi will be sum of all smaller d_vf_aoi values
        total_vf_aoi = faoi_integrand.sum(axis=1)
        # Make sure vf is counted as zero if the wedge is super small
        total_vf_aoi = np.where(
            np.abs(high_angles - low_angles) < DISTANCE_TOLERANCE, 0.,
            total_vf_aoi)

        return total_vf_aoi

    def _calculate_vfaoi_integrand(self, low_angles, high_angles,
                                   is_back=True):
        """
        Calculate the timeseries view factors with aoi loss integrand
        given the low and high angles that define the surface.

        Parameters
        ----------
        low_angles : np.ndarray
            Low AOI angles (between 0 and 180 deg), length = n_timestamps
        high_angles : np.ndarray
            High AOI angles (between 0 and 180 deg), length = n_timestamps.
            Should be bigger than ``low_angles``
        is_back : bool
            Flag specifying whether pv row surface is on back or front side
            of PV row (Default = True)
        Returns
        -------
        np.ndarray
            vf_aoi integrand values for all timestamps
            shape = (n_timestamps, n_integral_sections)
        """

        # Turn into dimension: [n_timestamps, n_integral_sections]
        low_angles_mat = np.tile(low_angles, (self.n_integral_sections, 1)).T
        high_angles_mat = np.tile(high_angles, (self.n_integral_sections, 1)).T

        # Filter out integrand values outside of range
        include_integral_section = ((low_angles_mat <= self.aoi_angles_high) &
                                    (high_angles_mat > self.aoi_angles_low))
        # The integrand values are different for front and back sides
        if is_back:
            faoi_integrand = np.where(include_integral_section,
                                      self.integrand_back, 0.)
        else:
            faoi_integrand = np.where(include_integral_section,
                                      self.integrand_front, 0.)

        return faoi_integrand

    @staticmethod
    def _calculate_aoi_angles(u_vector, centroid, point):
        """Calculate AOI angles from direction vector of surface,
        centroid point of that surface, and point from another surface

        Parameters
        ----------
        u_vector : np.ndarray
            Direction vector of the surface for which to calculate AOI angles
        centroid : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Centroid point of surface for which to calculate AOI angles
        point : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Point of remote surface that will determine AOI angle

        Returns
        -------
        np.ndarray
            AOI angles formed by remote point and centroid on surface,
            measured against surface direction vector [degrees]
        """
        v_vector = np.array([point.x - centroid.x, point.y - centroid.y])
        dot_product = u_vector[0, :] * v_vector[0, :] \
            + u_vector[1, :] * v_vector[1, :]
        u_norm = np.linalg.norm(u_vector, axis=0)
        v_norm = np.linalg.norm(v_vector, axis=0)
        cos_theta = dot_product / (u_norm * v_norm)
        # because of round off errors, cos_theta can be slightly > 1,
        # or slightly < -1, so clip it
        cos_theta = np.clip(cos_theta, -1., 1.)
        aoi_angles = np.rad2deg(np.arccos(cos_theta))
        return aoi_angles

    @staticmethod
    def _vf(aoi_1, aoi_2):
        """Calculate view factor from infinitesimal surface to infinite band.

        See illustration: http://www.thermalradiation.net/sectionb/B-71.html
        Here we're using angles measured from the horizontal

        Parameters
        ----------
        aoi_1 : np.ndarray
            Lower angles defining the infinite band
        aoi_2 : np.ndarray
            Higher angles defining the infinite band

        Returns
        -------
        np.ndarray
            View factors from infinitesimal surface to infinite strip

        """
        return 0.5 * np.abs(cosd(aoi_1) - cosd(aoi_2))

    @staticmethod
    def _angle_with_x_axis(pt_1, pt_2):
        """Angle with x-axis of vector going from pt_1 to pt_2

        Parameters
        ----------
        pt_1 : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries point coordinates of point 1
        pt_2 : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries point coordinates of point 2

        Returns
        -------
        np.ndarray
            Angle between vector pt_1->pt_2 and x-axis
        """
        return np.arctan2(pt_2.y - pt_1.y, pt_2.x - pt_1.x)

    def rho_from_faoi_fn(self, is_back):
        """Calculate global average reflectivity from faoi function
        for either side of the PV row (requires calculating view factors)

        Parameters
        ----------
        is_back : bool
            Flag specifying whether to use front or back faoi function
        Returns
        -------
        rho_average : float
            Global average reflectivity value of surface
        """
        # Will use x values at the middle of the integral sections
        aoi_angles = np.linspace(0., 180., num=self.n_integral_sections + 1)
        # Assumes that at least 2 aoi angle values, otherwise what's the point
        self.interval = aoi_angles[1] - aoi_angles[0]
        # Get integral intervals' low, high, and middle points
        aoi_angles_low = aoi_angles[:-1]
        aoi_angles_high = aoi_angles_low + self.interval
        aoi_angles_middle = aoi_angles_low + self.interval / 2.
        # Calculate faoi values using middle points of integral intervals
        if is_back:
            faoi_values = self.faoi_fn_back(aoi_angles_middle)
        else:
            faoi_values = self.faoi_fn_front(aoi_angles_middle)
        # Calculate small view factor values for each section
        vf_values = self._vf(aoi_angles_low, aoi_angles_high)
        # Multiply to get integrand
        integrand_values = faoi_values * vf_values
        return (1. - integrand_values.sum())


def faoi_fn_from_pvlib_sandia(pvmodule_name):
    """Generate a faoi function from a pvlib sandia PV module name

    Parameters
    ----------
    pvmodule_name : str
        Name of PV module in pvlib Sandia module database

    Returns
    -------
    faoi_function
        Function that returns positive loss values for numeric inputs
        between 0 and 180 degrees.
    """
    # Get Sandia module database from pvlib
    sandia_modules = pvlib.pvsystem.retrieve_sam('SandiaMod')
    # Grab pv module sandia coeffs from database
    pvmodule = sandia_modules[pvmodule_name]

    def fn(angles):
        """fAOI loss funtion: calculate how much light is absorbed at given
        incidence angles

        Parameters
        ----------
        angles : np.ndarray or list
            Angles measured from surface horizontal, from 0 de 180 deg

        Returns
        -------
        np.ndarray
            fAOI values
        """
        angles = np.array(angles) if isinstance(angles, list) else angles
        # Transform the inputs for the SAPM function
        angles = np.where(angles >= 90, angles - 90, 90. - angles)
        # Use pvlib sapm aoi loss method
        return pvlib.pvsystem.sapm_aoi_loss(angles, pvmodule, upper=1.)

    return fn



"""Module containing the functions to run engine calculations in normal or
parallel mode."""

from pvfactors import logging
from pvfactors.geometry import OrderedPVArray
from pvfactors.irradiance import HybridPerezOrdered
from pvfactors.viewfactors import VFCalculator
from pvfactors.engine import PVEngine
from multiprocessing import Pool, cpu_count
import numpy as np
import pandas as pd
from time import time


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def run_timeseries_engine(fn_build_report, pvarray_parameters,
                          timestamps, dni, dhi, solar_zenith, solar_azimuth,
                          surface_tilt, surface_azimuth, albedo,
                          cls_pvarray=OrderedPVArray, cls_engine=PVEngine,
                          cls_irradiance=HybridPerezOrdered,
                          cls_vf=VFCalculator,
                          fast_mode_pvrow_index=None,
                          fast_mode_segment_index=None,
                          irradiance_model_params=None,
                          vf_calculator_params=None,
                          ghi=None):
    """Run timeseries simulation without multiprocessing. This is the
    functional approach to the :py:class:`~pvfactors.engine.PVEngine` class.

    Parameters
    ----------
    fn_build_report : function
        Function that will build the report of the simulation
    pvarray_parameters : dict
        The parameters defining the PV array
    timestamps : array-like
        List of timestamps of the simulation.
    dni : array-like
        Direct normal irradiance values [W/m2]
    dhi : array-like
        Diffuse horizontal irradiance values [W/m2]
    solar_zenith : array-like
        Solar zenith angles [deg]
    solar_azimuth : array-like
        Solar azimuth angles [deg]
    surface_tilt : array-like
        Surface tilt angles, from 0 to 180 [deg]
    surface_azimuth : array-like
        Surface azimuth angles [deg]
    albedo : array-like
        Albedo values (or ground reflectivity)
    cls_pvarray : class of PV array, optional
        Class that will be used to build the PV array
        (Default =
        :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray` class)
    cls_engine : class of PV engine, optional
        Class of the engine to use to run the simulations (Default =
        :py:class:`~pvfactors.engine.PVEngine` class)
    cls_irradiance : class of irradiance model, optional
        The irradiance model that will be applied to the PV array
        (Default =
        :py:class:`~pvfactors.irradiance.models.HybridPerezOrdered` class)
    cls_vf : class of VF calculator, optional
        Calculator that will be used to calculate the view factor matrices
        (Default =
        :py:class:`~pvfactors.viewfactors.calculator.VFCalculator` class)
    fast_mode_pvrow_index : int, optional
        If a valid pvrow index is passed, then the PVEngine fast mode
        will be activated and the engine calculation will be done only
        for the back surface of the selected pvrow (Default = None)
    fast_mode_segment_index : int, optional
        If a segment index is passed, then the PVEngine fast mode
        will calculate back surface irradiance only for the
        selected segment of the selected back surface (Default = None)
    irradiance_model_params : dict, optional
        Dictionary of parameters that will be passed to the irradiance model
        class as kwargs at instantiation (Default = None)
    vf_calculator_params : dict, optional
        Dictionary of parameters that will be passed to the VF calculator
        class as kwargs at instantiation (Default = None)
    ghi : array-like, optional
        Global horizontal irradiance values [W/m2] (Default = None)

    Returns
    -------
    report
        Saved results from the simulation, as specified by user's report
        function
    """

    # Prepare input parameters
    irradiance_model_params = irradiance_model_params or {}
    vf_calculator_params = vf_calculator_params or {}
    # Instantiate classes and engine
    irradiance_model = cls_irradiance(**irradiance_model_params)
    vf_calculator = cls_vf(**vf_calculator_params)
    pvarray = cls_pvarray.init_from_dict(pvarray_parameters)
    eng = cls_engine(pvarray, irradiance_model=irradiance_model,
                     vf_calculator=vf_calculator,
                     fast_mode_pvrow_index=fast_mode_pvrow_index,
                     fast_mode_segment_index=fast_mode_segment_index)
    # Fit engine
    eng.fit(timestamps, dni, dhi, solar_zenith, solar_azimuth, surface_tilt,
            surface_azimuth, albedo, ghi=ghi)

    # Run all timesteps
    report = (eng.run_full_mode(fn_build_report=fn_build_report)
              if fast_mode_pvrow_index is None
              else eng.run_fast_mode(fn_build_report=fn_build_report))

    return report


def run_parallel_engine(report_builder, pvarray_parameters,
                        timestamps, dni, dhi, solar_zenith, solar_azimuth,
                        surface_tilt, surface_azimuth, albedo,
                        cls_pvarray=OrderedPVArray, cls_engine=PVEngine,
                        cls_irradiance=HybridPerezOrdered,
                        cls_vf=VFCalculator, fast_mode_pvrow_index=None,
                        fast_mode_segment_index=None,
                        irradiance_model_params=None,
                        vf_calculator_params=None, n_processes=2,
                        ghi=None):
    """Run timeseries simulation using multiprocessing. Here, instead of a
    function that will build the report, the users will need to pass a class
    (or an object).

    Parameters
    ----------
    report_builder : class or object
        Class or object that will build and merge the reports. It **must**
        have a ``build()`` and a ``merge()`` method that perform the tasks
    pvarray_parameters : dict
        The parameters defining the PV array
    timestamps : array-like
        List of timestamps of the simulation.
    dni : array-like
        Direct normal irradiance values [W/m2]
    dhi : array-like
        Diffuse horizontal irradiance values [W/m2]
    solar_zenith : array-like
        Solar zenith angles [deg]
    solar_azimuth : array-like
        Solar azimuth angles [deg]
    surface_tilt : array-like
        Surface tilt angles, from 0 to 180 [deg]
    surface_azimuth : array-like
        Surface azimuth angles [deg]
    albedo : array-like
        Albedo values (or ground reflectivity)
    cls_pvarray : class of PV array, optional
        Class that will be used to build the PV array
        (Default =
        :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray` class)
    cls_engine : class of PV engine, optional
        Class of the engine to use to run the simulations (Default =
        :py:class:`~pvfactors.engine.PVEngine` class)
    cls_irradiance : class of irradiance model, optional
        The irradiance model that will be applied to the PV array
        (Default =
        :py:class:`~pvfactors.irradiance.models.HybridPerezOrdered` class)
    cls_vf : class of VF calculator, optional
        Calculator that will be used to calculate the view factor matrices
        (Default =
        :py:class:`~pvfactors.viewfactors.calculator.VFCalculator` class)
    fast_mode_pvrow_index : int, optional
        If a valid pvrow index is passed, then the PVEngine fast mode
        will be activated and the engine calculation will be done only
        for the back surface of the selected pvrow (Default = None)
    fast_mode_segment_index : int, optional
        If a segment index is passed, then the PVEngine fast mode
        will calculate back surface irradiance only for the
        selected segment of the selected back surface (Default = None)
    irradiance_model_params : dict, optional
        Dictionary of parameters that will be passed to the irradiance model
        class as kwargs at instantiation (Default = None)
    vf_calculator_params : dict, optional
        Dictionary of parameters that will be passed to the VF calculator
        class as kwargs at instantiation (Default = None)
    n_processes : int, optional
        Number of parallel processes to run for the calculation (Default = 2)
    ghi : array-like, optional
        Global horizontal irradiance values [W/m2] (Default = None)

    Returns
    -------
    report
        Saved results from the simulation, as specified by user's report
        class (or object)
    """
    # Choose number of workers
    if n_processes == -1:
        n_processes = cpu_count()

    # Make sure albedo is iterable
    if np.isscalar(albedo):
        albedo = albedo * np.ones(len(dni))

    # Prepare class input parameters
    irradiance_model_params = irradiance_model_params or {}
    vf_calculator_params = vf_calculator_params or {}

    # Fix: np.array_split doesn't work well on pd.DatetimeIndex objects
    if isinstance(timestamps, pd.DatetimeIndex):
        timestamps = timestamps.to_pydatetime()

    # Split all arguments according to number of processes
    (folds_timestamps, folds_surface_azimuth, folds_surface_tilt,
     folds_solar_zenith, folds_solar_azimuth, folds_dni, folds_dhi,
     folds_albedo) = map(np.array_split,
                         [timestamps, surface_azimuth, surface_tilt,
                          solar_zenith, solar_azimuth, dni, dhi, albedo],
                         [n_processes] * 8)
    folds_report_builder = [report_builder] * n_processes
    folds_params = [pvarray_parameters] * n_processes
    folds_cls_pvarray = [cls_pvarray] * n_processes
    folds_cls_engine = [cls_engine] * n_processes
    folds_cls_irradiance = [cls_irradiance] * n_processes
    folds_cls_vf = [cls_vf] * n_processes
    folds_fast_mode_pvrow_index = [fast_mode_pvrow_index] * n_processes
    folds_fast_mode_segment_index = [fast_mode_segment_index] * n_processes
    folds_irradiance_model_params = [irradiance_model_params] * n_processes
    folds_vf_calculator_params = [vf_calculator_params] * n_processes
    folds_ghi = ([ghi] * n_processes if ghi is None else
                 np.array_split(ghi, n_processes))
    report_indices = list(range(n_processes))

    # Zip all the folds together
    list_folds = zip(*(folds_report_builder, folds_params, folds_timestamps,
                       folds_dni, folds_dhi, folds_solar_zenith,
                       folds_solar_azimuth, folds_surface_tilt,
                       folds_surface_azimuth, folds_albedo, folds_cls_pvarray,
                       folds_cls_engine, folds_cls_irradiance, folds_cls_vf,
                       folds_fast_mode_pvrow_index,
                       folds_fast_mode_segment_index,
                       folds_irradiance_model_params,
                       folds_vf_calculator_params, folds_ghi,
                       report_indices))

    # Start multiprocessing
    pool = Pool(n_processes)
    start = time()
    indexed_reports = pool.map(_run_serially, list_folds)
    end = time()
    pool.close()
    pool.join()
    sorted_indexed_reports = sorted(indexed_reports, key=lambda tup: tup[1])
    sorted_reports = [tup[0] for tup in sorted_indexed_reports]

    LOGGER.info("Parallel calculation elapsed time: %s sec" % str(end - start))

    report = report_builder.merge(sorted_reports)

    return report


# Utility function for parallel run
def _run_serially(args):
    """Helper function used to run calculations in parallel

    Parameters
    ----------
    args : tuple
        List of arguments where most will be used in
        :py:function:`~pvfactors.run.run_timeseries_engine`

    Returns
    -------
    report
        Saved results from the simulation, as specified by user's report
        class (or object)
    idx : int
        Index of the report, which will be used to sort the final list of
        reports after all the parallel simulations are over

    """
    report_builder, pvarray_parameters, timestamps, dni, dhi, \
        solar_zenith, solar_azimuth, surface_tilt, surface_azimuth,\
        albedo, cls_pvarray, cls_engine, cls_irradiance, cls_vf, \
        fast_mode_pvrow_index, fast_mode_segment_index, \
        irradiance_model_params, vf_calculator_params, ghi, idx = args

    report = run_timeseries_engine(
        report_builder.build, pvarray_parameters,
        timestamps, dni, dhi, solar_zenith, solar_azimuth,
        surface_tilt, surface_azimuth, albedo,
        cls_pvarray=cls_pvarray, cls_engine=cls_engine,
        cls_irradiance=cls_irradiance, cls_vf=cls_vf,
        fast_mode_pvrow_index=fast_mode_pvrow_index,
        fast_mode_segment_index=fast_mode_segment_index,
        irradiance_model_params=irradiance_model_params,
        vf_calculator_params=vf_calculator_params,
        ghi=ghi)

    return report, idx



"""Module with classes and functions to calculate views and view factors"""

from pvfactors.config import DISTANCE_TOLERANCE
from pvfactors.viewfactors.vfmethods import VFTsMethods
from pvfactors.viewfactors.aoimethods import AOIMethods
import numpy as np


class VFCalculator(object):
    """This calculator class will be used for the calculation of view factors
    for :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`, and it will
    rely on both :py:class:`~pvfactors.viewfactors.vfmethods.VFTsMethods`
    and :py:class:`~pvfactors.viewfactors.aoimethods.AOIMethods`"""

    def __init__(self, faoi_fn_front=None, faoi_fn_back=None,
                 n_aoi_integral_sections=300):
        """Initialize the view factor calculator with the calculation methods
        that will be used. The AOI methods will not be instantiated if an
        fAOI function is missing.

        Parameters
        ----------
        faoi_fn_front : function or object, optional
            Function (or object containing ``faoi`` method)
            which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the front side of PV rows (default = None)
        faoi_fn_back : function or object, optional
            Function (or object containing ``faoi`` method)
            which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the back side of PV rows (default = None)
        n_integral_sections : int, optional
            Number of integral divisions of the 0 to 180 deg interval
            to use for the fAOI loss integral (default = 300)
        """
        self.vf_ts_methods = VFTsMethods()
        # Do not instantiate AOIMethods if missing faoi function
        if (faoi_fn_front is None) or (faoi_fn_back is None):
            self.vf_aoi_methods = None
        else:
            # Check whether got function or object, and take ``faoi`` method
            # if object was passed
            faoi_fn_front = (faoi_fn_front.faoi
                             if hasattr(faoi_fn_front, 'faoi')
                             else faoi_fn_front)
            faoi_fn_back = (faoi_fn_back.faoi
                            if hasattr(faoi_fn_back, 'faoi') else faoi_fn_back)
            self.vf_aoi_methods = AOIMethods(
                faoi_fn_front, faoi_fn_back,
                n_integral_sections=n_aoi_integral_sections)
        # Saved matrices
        self.vf_matrix = None
        self.vf_aoi_matrix = None

    def fit(self, n_timestamps):
        """Fit the view factor calculator to the timeseries inputs.

        Parameters
        ----------
        n_timestamps : int
            Number of simulation timestamps
        """
        if self.vf_aoi_methods is not None:
            self.vf_aoi_methods.fit(n_timestamps)

    def build_ts_vf_matrix(self, pvarray):
        """Calculate timeseries view factor matrix for the given
        ordered pv array

        Parameters
        ----------
        pvarray : :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`
            PV array whose timeseries view factor matrix to calculate

        Returns
        -------
        np.ndarray
            Timeseries view factor matrix, with 3 dimensions:
            [n_surfaces, n_surfaces, n_timesteps]
        """

        # Initialize matrix
        rotation_vec = pvarray.rotation_vec
        tilted_to_left = rotation_vec > 0
        n_steps = len(rotation_vec)
        n_ts_surfaces = pvarray.n_ts_surfaces
        vf_matrix = np.zeros((n_ts_surfaces + 1, n_ts_surfaces + 1, n_steps),
                             dtype=float)  # don't forget to include the sky

        # Get timeseries objects
        ts_ground = pvarray.ts_ground
        ts_pvrows = pvarray.ts_pvrows

        # Calculate ts view factors between pvrow and ground surfaces
        self.vf_ts_methods.vf_pvrow_gnd_surf(ts_pvrows, ts_ground,
                                             tilted_to_left, vf_matrix)
        # Calculate view factors between pv rows
        self.vf_ts_methods.vf_pvrow_to_pvrow(ts_pvrows, tilted_to_left,
                                             vf_matrix)
        # Calculate view factors to sky
        vf_matrix[:-1, -1, :] = 1. - np.sum(vf_matrix[:-1, :-1, :], axis=1)
        # This is not completely accurate yet, we need to set the sky vf
        # to zero when the surfaces have zero length
        for i, ts_surf in enumerate(pvarray.all_ts_surfaces):
            vf_matrix[i, -1, :] = np.where(ts_surf.length > DISTANCE_TOLERANCE,
                                           vf_matrix[i, -1, :], 0.)

        # Save in calculator
        self.vf_matrix = vf_matrix

        return vf_matrix

    def build_ts_vf_aoi_matrix(self, pvarray, rho_mat):
        """Calculate the view factor aoi matrix elements from all PV row
        surfaces to all other surfaces, only.
        If the AOI methods are available, the vf_aoi_matrix will account
        for reflection losses that are AOI specific. Otherwise it will
        assume that all the reflection losses are diffuse.

        Notes
        -----
        When using fAOI methods, this will not calculate
        view factors from ground surfaces to PV row surfaces, so the users
        will need to run
        :py:meth:`~pvfactors.viewfactors.calculator.VFCalculator.build_ts_vf_matrix`
        first if they want the complete matrix, otherwise those entries will
        have zero values in them.


        Parameters
        ----------
        pvarray : :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`
            PV array whose timeseries view factor AOI matrix to calculate
        rho_mat : np.ndarray
            2D matrix of reflectivity values for all the surfaces in the
            PV array + sky.
            Shape = [n_ts_surfaces + 1, n_ts_surfaces + 1, n_timestamps]

        Returns
        -------
        np.ndarray
            Timeseries view factor matrix for infinitesimal PV row surfaces,
            and accounting for AOI losses, with 3 dimensions:
            [n_surfaces, n_surfaces, n_timesteps]
        """
        # Initialize matrix
        rotation_vec = pvarray.rotation_vec
        tilted_to_left = rotation_vec > 0
        n_steps = len(rotation_vec)
        n_ts_surfaces = pvarray.n_ts_surfaces
        vf_aoi_matrix = np.zeros(
            (n_ts_surfaces + 1, n_ts_surfaces + 1, n_steps),
            dtype=float) if self.vf_matrix is None else self.vf_matrix

        # Get timeseries objects
        ts_ground = pvarray.ts_ground
        ts_pvrows = pvarray.ts_pvrows

        if self.vf_aoi_methods is None:
            # The reflection losses will be considered all diffuse.
            faoi_diffuse = 1. - rho_mat
            vf_aoi_matrix = faoi_diffuse * vf_aoi_matrix
        else:
            # Calculate vf_aoi between pvrow and ground surfaces
            self.vf_aoi_methods.vf_aoi_pvrow_to_gnd(ts_pvrows, ts_ground,
                                                    tilted_to_left,
                                                    vf_aoi_matrix)
            # Calculate vf_aoi between pvrows
            self.vf_aoi_methods.vf_aoi_pvrow_to_pvrow(
                ts_pvrows, tilted_to_left, vf_aoi_matrix)
            # Calculate vf_aoi between prows and sky
            self.vf_aoi_methods.vf_aoi_pvrow_to_sky(
                ts_pvrows, ts_ground, tilted_to_left, vf_aoi_matrix)

        # Save results
        self.vf_aoi_matrix = vf_aoi_matrix

        return vf_aoi_matrix

    def get_vf_ts_pvrow_element(self, pvrow_idx, pvrow_element, ts_pvrows,
                                ts_ground, rotation_vec, pvrow_width):
        """Calculate timeseries view factors of timeseries pvrow element
        (segment or surface) to all other elements of the PV array.

        Parameters
        ----------
        pvrow_idx : int
            Index of the timeseries PV row for which we want to calculate the
            back surface irradiance
        pvrow_element : \
        :py:class:`~pvfactors.geometry.timeseries.TsDualSegment` \
        or :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            Timeseries PV row element for which to calculate view factors
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        ts_ground : :py:class:`~pvfactors.geometry.timeseries.TsGround`
            Timeseries ground of the PV array
        rotation_vec : np.ndarray
            Timeseries rotation vector of the PV rows in [deg]
        pvrow_width : float
            Width of the timeseries PV rows in the array in [m]

        Returns
        -------
        view_factors : dict
            Dictionary of the timeseries view factors to all types of surfaces
            in the PV array. List of keys include: 'to_each_gnd_shadow',
            'to_gnd_shaded', 'to_gnd_illum', 'to_gnd_total', 'to_pvrow_total',
            'to_pvrow_shaded', 'to_pvrow_illum', 'to_sky'
        """
        tilted_to_left = rotation_vec > 0
        n_shadows = len(ts_pvrows)
        n_steps = len(rotation_vec)
        pvrow_element_coords = pvrow_element.coords
        pvrow_element_length = pvrow_element_coords.length

        # Get shadows on left and right sides of PV row
        shadows_coords_left = \
            ts_ground.shadow_coords_left_of_cut_point(pvrow_idx)
        shadows_coords_right = \
            ts_ground.shadow_coords_right_of_cut_point(pvrow_idx)
        # Calculate view factors to ground shadows
        list_vf_to_obstructed_gnd_shadows = []
        for i in range(n_shadows):
            shadow_left = shadows_coords_left[i]
            shadow_right = shadows_coords_right[i]
            # vfs to obstructed gnd shadows
            vf_obstructed_shadow = (
                self.vf_ts_methods.calculate_vf_to_shadow_obstruction_hottel(
                    pvrow_element, pvrow_idx, n_shadows, n_steps,
                    tilted_to_left, ts_pvrows, shadow_left, shadow_right,
                    pvrow_element_length))
            list_vf_to_obstructed_gnd_shadows.append(vf_obstructed_shadow)
        list_vf_to_obstructed_gnd_shadows = np.array(
            list_vf_to_obstructed_gnd_shadows)

        # Calculate view factors to shaded ground
        vf_shaded_gnd = np.sum(list_vf_to_obstructed_gnd_shadows, axis=0)

        # Calculate view factors to whole ground
        vf_gnd_total = self.vf_ts_methods.calculate_vf_to_gnd(
            pvrow_element_coords, pvrow_idx, n_shadows, n_steps,
            ts_ground.y_ground, ts_ground.cut_point_coords[pvrow_idx],
            pvrow_element_length, tilted_to_left, ts_pvrows)

        # Calculate view factors to illuminated ground
        vf_illum_gnd = vf_gnd_total - vf_shaded_gnd

        # Calculate view factors to pv rows
        vf_pvrow_total, vf_pvrow_shaded = \
            self.vf_ts_methods.calculate_vf_to_pvrow(
                pvrow_element_coords, pvrow_idx, n_shadows, n_steps, ts_pvrows,
                pvrow_element_length, tilted_to_left, pvrow_width,
                rotation_vec)
        vf_pvrow_illum = vf_pvrow_total - vf_pvrow_shaded

        # Calculate view factors to sky
        vf_to_sky = 1. - vf_gnd_total - vf_pvrow_total

        # return all timeseries view factors
        view_factors = {
            'to_each_gnd_shadow': list_vf_to_obstructed_gnd_shadows,
            'to_gnd_shaded': vf_shaded_gnd,
            'to_gnd_illum': vf_illum_gnd,
            'to_gnd_total': vf_gnd_total,
            'to_pvrow_total': vf_pvrow_total,
            'to_pvrow_shaded': vf_pvrow_shaded,
            'to_pvrow_illum': vf_pvrow_illum,
            'to_sky': vf_to_sky
        }

        return view_factors



"""Module containing specific irradiance modeling functions and tools.
More work probably needs to be done to improve the computation speed here, and
in theory a big part of ``perez_diffuse_luminance`` should be handled by pvlib
"""
import numpy as np
import pandas as pd
from pvlib import atmosphere, irradiance
from pvlib.tools import cosd, sind
import math
from pvfactors import PVFactorsError
from pvfactors.config import \
    SIGMA, N_SIGMA, GAUSSIAN_DIAMETER_CIRCUMSOLAR, RADIUS_CIRCUMSOLAR


def perez_diffuse_luminance(timestamps, surface_tilt, surface_azimuth,
                            solar_zenith, solar_azimuth, dni, dhi):
    """Function used to calculate the luminance and the view factor terms from the
    Perez diffuse light transposition model, as implemented in the
    ``pvlib-python`` library.
    This function was custom made to allow the calculation of the circumsolar
    component on the back surface as well. Otherwise, the ``pvlib``
    implementation would ignore it.

    Parameters
    ----------
    timestamps : array-like
        simulation timestamps
    surface_tilt : array-like
        Surface tilt angles in decimal degrees.
        surface_tilt must be >=0 and <=180.
        The tilt angle is defined as degrees from horizontal
        (e.g. surface facing up = 0, surface facing horizon = 90)
    surface_azimuth : array-like
        The azimuth of the rotated panel,
        determined by projecting the vector normal to the panel's surface to
        the earth's surface [degrees].
    solar_zenith : array-like
        solar zenith angles
    solar_azimuth : array-like
        solar azimuth angles
    dni : array-like
        values for direct normal irradiance
    dhi : array-like
        values for diffuse horizontal irradiance

    Returns
    -------
    df_inputs : `pandas.DataFrame`
        Dataframe with the following columns:
        ['solar_zenith', 'solar_azimuth', 'surface_tilt', 'surface_azimuth',
        'dhi', 'dni', 'vf_horizon', 'vf_circumsolar', 'vf_isotropic',
        'luminance_horizon', 'luminance_circuqmsolar', 'luminance_isotropic',
        'poa_isotropic', 'poa_circumsolar', 'poa_horizon', 'poa_total_diffuse']

    """
    # Create a dataframe to help filtering on all arrays
    df_inputs = pd.DataFrame(
        {'surface_tilt': surface_tilt, 'surface_azimuth': surface_azimuth,
         'solar_zenith': solar_zenith, 'solar_azimuth': solar_azimuth,
         'dni': dni, 'dhi': dhi},
        index=pd.DatetimeIndex(timestamps))

    dni_et = irradiance.get_extra_radiation(df_inputs.index.dayofyear)
    am = atmosphere.get_relative_airmass(df_inputs.solar_zenith)

    # Need to treat the case when the sun is hitting the back surface of pvrow
    aoi_proj = irradiance.aoi_projection(
        df_inputs.surface_tilt, df_inputs.surface_azimuth,
        df_inputs.solar_zenith, df_inputs.solar_azimuth)
    sun_hitting_back_surface = ((aoi_proj < 0) &
                                (df_inputs.solar_zenith <= 90))
    df_inputs_back_surface = df_inputs.loc[sun_hitting_back_surface].copy()
    # Reverse the surface normal to switch to back-surface circumsolar calc
    df_inputs_back_surface.loc[:, 'surface_azimuth'] = (
        df_inputs_back_surface.loc[:, 'surface_azimuth'] - 180.)
    df_inputs_back_surface.loc[:, 'surface_azimuth'] = np.mod(
        df_inputs_back_surface.loc[:, 'surface_azimuth'], 360.
    )
    df_inputs_back_surface.loc[:, 'surface_tilt'] = (
        180. - df_inputs_back_surface.surface_tilt)

    if df_inputs_back_surface.shape[0] > 0:
        # Use recursion to calculate circumsolar luminance for back surface
        df_inputs_back_surface = perez_diffuse_luminance(
            *breakup_df_inputs(df_inputs_back_surface))

    # Calculate Perez diffuse components
    components = irradiance.perez(df_inputs.surface_tilt,
                                  df_inputs.surface_azimuth,
                                  df_inputs.dhi, df_inputs.dni,
                                  dni_et,
                                  df_inputs.solar_zenith,
                                  df_inputs.solar_azimuth,
                                  am,
                                  return_components=True)

    # Calculate Perez view factors:
    a = irradiance.aoi_projection(
        df_inputs.surface_tilt,
        df_inputs.surface_azimuth, df_inputs.solar_zenith,
        df_inputs.solar_azimuth)
    a = np.maximum(a, 0)
    b = cosd(df_inputs.solar_zenith)
    b = np.maximum(b, cosd(85))

    vf_perez = pd.DataFrame({
        'vf_horizon': sind(df_inputs.surface_tilt),
        'vf_circumsolar': a / b,
        'vf_isotropic': (1. + cosd(df_inputs.surface_tilt)) / 2.
    },
        index=df_inputs.index
    )

    # Calculate diffuse luminance
    luminance = pd.DataFrame(
        np.array([
            components['horizon'] / vf_perez['vf_horizon'],
            components['circumsolar'] / vf_perez['vf_circumsolar'],
            components['isotropic'] / vf_perez['vf_isotropic']
        ]).T,
        index=df_inputs.index,
        columns=['luminance_horizon', 'luminance_circumsolar',
                 'luminance_isotropic']
    )
    luminance.loc[components['sky_diffuse'] == 0, :] = 0.

    # Format components column names
    components = components.rename(columns={'isotropic': 'poa_isotropic',
                                            'circumsolar': 'poa_circumsolar',
                                            'horizon': 'poa_horizon'})

    df_inputs = pd.concat([df_inputs, components, vf_perez, luminance],
                          axis=1, join='outer')
    df_inputs = df_inputs.rename(columns={'sky_diffuse': 'poa_total_diffuse'})

    # Adjust the circumsolar luminance when it hits the back surface
    if df_inputs_back_surface.shape[0] > 0:
        df_inputs.loc[sun_hitting_back_surface, 'luminance_circumsolar'] = (
            df_inputs_back_surface.loc[:, 'luminance_circumsolar'])

    return df_inputs


def breakup_df_inputs(df_inputs):
    """Helper function: It is sometimes easier to provide a dataframe than a list
    of arrays: this function does the job of breaking up the dataframe into a
    list of expected 1-dim arrays

    Parameters
    ----------
    df_inputs : ``pandas.DataFrame``
        timestamp-indexed dataframe with following columns:
        'surface_azimuth', 'surface_tilt', 'solar_zenith', 'solar_azimuth',
        'dni', 'dhi'

    Returns
    -------
    all 1-dim ``numpy.ndarray``
        ``timestamps``, ``tilt_angles``, ``surface_azimuth``,
        ``solar_zenith``, ``solar_azimuth``, ``dni``, ``dhi``

    """
    timestamps = pd.to_datetime(df_inputs.index)
    surface_azimuth = df_inputs.surface_azimuth.values
    surface_tilt = df_inputs.surface_tilt.values
    solar_zenith = df_inputs.solar_zenith.values
    solar_azimuth = df_inputs.solar_azimuth.values
    dni = df_inputs.dni.values
    dhi = df_inputs.dhi.values

    return (timestamps, surface_tilt, surface_azimuth,
            solar_zenith, solar_azimuth, dni, dhi)


def calculate_circumsolar_shading(percentage_distance_covered,
                                  model='uniform_disk'):
    """Select the model to calculate circumsolar shading based on the current PV
    array condition.

    Parameters
    ----------
    percentage_distance_covered : float
        this represents how much of the
        circumsolar diameter is covered by the neighboring row [in %]
    model : str, optional
        name of the circumsolar shading model to use:
        'uniform_disk' and 'gaussian' are the two models currently available
        (Default value = 'uniform_disk')

    Returns
    -------
    float
        shading percentage of circumsolar area

    """
    if model == 'uniform_disk':
        perc_shading = uniform_circumsolar_disk_shading(
            percentage_distance_covered)

    elif model == 'gaussian':
        perc_shading = gaussian_shading(percentage_distance_covered)

    else:
        raise PVFactorsError(
            'calculate_circumsolar_shading: model does not exist: ' +
            '%s' % model)

    return perc_shading


def integral_default_gaussian(y, x):
    """Calculate the value of the integral from x to y of the erf function

    Parameters
    ----------
    y : float
        upper limit
    x : float
        lower limit

    Returns
    -------
    float
        Calculated value of integral

    """
    return 0.5 * (math.erf(x) - math.erf(y))


def gaussian_shading(percentage_distance_covered):
    """Calculate circumsolar shading by assuming that the irradiance profile on
    a 2D section of the circumsolar disk is Gaussian

    Parameters
    ----------
    percentage_distance_covered : float
        [in %], proportion of the
        circumsolar disk area covered by the neighboring pvrow

    Returns
    -------
    float
        percentage shading in terms of irradiance (using gaussian profile)

    """
    if percentage_distance_covered < 0.:
        perc_shading = 0.
    elif percentage_distance_covered > 100.:
        perc_shading = 100.
    else:
        y = - N_SIGMA * SIGMA
        x = (y +
             percentage_distance_covered / 100. *
             GAUSSIAN_DIAMETER_CIRCUMSOLAR)
        area = integral_default_gaussian(y, x)
        total_gaussian_area = integral_default_gaussian(- N_SIGMA * SIGMA,
                                                        N_SIGMA * SIGMA)
        perc_shading = area / total_gaussian_area * 100.

    return perc_shading


def gaussian(x, mu=0., sigma=1.):
    """Gaussian density function

    Parameters
    ----------
    x : float
        argument of function
    mu : float, optional
        mean of the gaussian (Default value = 0.)
    sigma : float, optional
        standard deviation of the gaussian (Default value = 1.)

    Returns
    -------
    float
        value of guassian function at point x

    """
    return (1. / (sigma * np.sqrt(2. * np.pi)) *
            np.exp(- 0.5 * np.power((x - mu) / sigma, 2)))


def uniform_circumsolar_disk_shading(percentage_distance_covered):
    """Calculate the percentage shading of circumsolar irradiance. This
    model considers circumsolar to be a disk, and calculates the
    percentage shaded based on the percentage "distance covered",
    which is how much of the disk's diameter is covered by the
    neighboring object.

    Parameters
    ----------
    percentage_distance_covered : float
        distance covered of circumsolar disk diameter
        [% - values from 0 to 100]

    Returns
    -------
    float
        value of guassian function at point x

    """

    # Define a circumsolar disk
    r_circumsolar = RADIUS_CIRCUMSOLAR
    d_circumsolar = 2 * r_circumsolar
    area_circumsolar = np.pi * r_circumsolar**2.

    # Calculate circumsolar on case by case
    distance_covered = percentage_distance_covered / 100. * d_circumsolar

    if distance_covered <= 0:
        # No shading of circumsolar
        percent_shading = 0.
    elif (distance_covered > 0.) & (distance_covered <= r_circumsolar):
        # Theta is the full circle sector angle (not half) used to
        # calculate circumsolar shading
        theta = 2 * np.arccos((r_circumsolar - distance_covered) /
                              r_circumsolar)  # rad
        area_circle_sector = 0.5 * r_circumsolar ** 2 * theta
        area_triangle_sector = 0.5 * r_circumsolar ** 2 * np.sin(theta)
        area_shaded = area_circle_sector - area_triangle_sector
        percent_shading = area_shaded / area_circumsolar * 100.

    elif (distance_covered > r_circumsolar) & (distance_covered <
                                               d_circumsolar):
        distance_uncovered = d_circumsolar - distance_covered

        # Theta is the full circle sector angle (not half) used to
        # calculate circumsolar shading
        theta = 2 * np.arccos((r_circumsolar - distance_uncovered) /
                              r_circumsolar)  # rad
        area_circle_sector = 0.5 * r_circumsolar ** 2 * theta
        area_triangle_sector = 0.5 * r_circumsolar ** 2 * np.sin(theta)
        area_not_shaded = area_circle_sector - area_triangle_sector
        percent_shading = (1. - area_not_shaded / area_circumsolar) * 100.

    else:
        # Total shading of circumsolar: distance_covered >= d_circumsolar
        percent_shading = 100.

    return percent_shading


def calculate_horizon_band_shading(shading_angle, horizon_band_angle):
    """Calculate the percentage shading of the horizon band; which is just the
    proportion of what's masked by the neighboring row since we assume
    uniform luminance values for the band

    Parameters
    ----------
    shading_angle : np.ndarray
        the elevation angle to use for shading
    horizon_band_angle : float
        the total elevation angle of the horizon band

    Returns
    -------
    float
        shading percentage of horizon band

    """
    percent_shading = np.where(shading_angle >= horizon_band_angle, 100., 0.)
    percent_shading = np.where(
        (shading_angle >= 0) & (shading_angle < horizon_band_angle),
        100. * shading_angle / horizon_band_angle, percent_shading)
    return percent_shading



""" Install pvfactors """

from setuptools import setup
import versioneer

DESCRIPTION = (
    '2D View Factor Model to calculate the irradiance incident on ' +
    'various surfaces of PV arrays')
with open('README.rst', 'r') as f:
    LONG_DESCRIPTION = f.read()
with open('requirements.txt', 'r') as f:
    INSTALL_REQUIRES = list(f)

DISTNAME = 'pvfactors'
AUTHOR = 'SunPower'
MAINTAINER_EMAIL = 'marc.abouanoma@sunpowercorp.com'
URL = 'https://github.com/SunPower/pvfactors'
PACKAGES = ['pvfactors', 'pvfactors.geometry', 'pvfactors.irradiance',
            'pvfactors.viewfactors']
LICENSE = 'BSD 3-Clause'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
]

TESTS_REQUIRES = ['pytest>=3.2.1', 'pytest-mock>=1.10.0', 'mock']

setup(name=DISTNAME,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      author=AUTHOR,
      maintainer_email=MAINTAINER_EMAIL,
      url=URL,
      packages=PACKAGES,
      classifiers=CLASSIFIERS,
      install_requires=INSTALL_REQUIRES,
      extras_require={
          'testing': TESTS_REQUIRES,
          'docs': ['Sphinx', 'sphinx_rtd_theme', 'nbsphinx',
                   'sphinxcontrib-napoleon', 'sphinxcontrib_github_alt',
                   'ipykernel']
      },
      license=LICENSE
      )
